#!/usr/bin/env python3

from trojanzoo.configs import config
from trojanzoo.datasets import Dataset
from trojanzoo.models import Model
from trojanzoo.attacks import Attack
from trojanzoo.utils.module import get_name, ModelProcess
from trojanzoo.utils.output import ansi

import os
from abc import ABC, abstractmethod

from typing import TYPE_CHECKING
from trojanzoo.configs import Config
import argparse    # TODO: python 3.10
if TYPE_CHECKING:
    pass


class Defense(ABC, ModelProcess):
    r"""
    | An abstract class representing a defense.
    | It inherits :class:`trojanzoo.utils.module.ModelProcess`.

    Note:
        This is the implementation of defense.
        For users, please use :func:`create` instead, which is more user-friendly.

    Attributes:
        attack (trojanzoo.attacks.Attack | None): The attack instance.
    """
    name: str = 'defense'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        r"""Add defense arguments to argument parser group.
        View source to see specific arguments.

        Note:
            This is the implementation of adding arguments.
            The concrete defense class may override this method to add more arguments.
            For users, please use :func:`add_argument()` instead, which is more user-friendly.
        """
        group.add_argument('--defense', dest='defense_name')
        group.add_argument('--defense_dir',
                           help='directory to contain defense results')
        return group

    def __init__(self, attack: Attack = None, **kwargs):
        super().__init__(**kwargs)
        self.attack = attack

    @abstractmethod
    def detect(self, *args, **kwargs):
        r"""Main detect method (need overriding)."""
        ...


def add_argument(parser: argparse.ArgumentParser,
                 defense_name: None | str = None,
                 defense: None | str | Defense = None,
                 class_dict: dict[str, type[Defense]] = {}):
    r"""
    | Add defense arguments to argument parser.
    | For specific arguments implementation, see :meth:`Defense.add_argument()`.

    Args:
        parser (argparse.ArgumentParser): The parser to add arguments.
        defense_name (str): The defense name.
        defense (str | Defense): The defense instance or defense name
            (as the alias of `defense_name`).
        class_dict (dict[str, type[Defense]]):
            Map from defense name to defense class.
            Defaults to ``{}``.

    Returns:
        argparse._ArgumentGroup: The argument group.
    """
    defense_name = get_name(
        name=defense_name, module=defense, arg_list=['--defense'])
    group = parser.add_argument_group(
        '{yellow}defense{reset}'.format(**ansi), description=defense_name)
    try:
        DefenseType = class_dict[defense_name]
    except KeyError:
        if defense_name is None:
            print(f'{ansi["red"]}you need to first claim the defense name '
                  f'using "--defense".{ansi["reset"]}')
        print(f'{defense_name} not in \n{list(class_dict.keys())}')
        raise
    return DefenseType.add_argument(group)


def create(defense_name: None | str = None, defense: None | str | Defense = None,
           folder_path: None | str = None,
           dataset_name: str = None, dataset: None | str | Dataset = None,
           model_name: str = None, model: None | str | Model = None,
           config: Config = config, class_dict: dict[str, type[Defense]] = {},
           **kwargs):
    r"""
    | Create a defense instance.
    | For arguments not included in :attr:`kwargs`,
      use the default values in :attr:`config`.
    | The default value of :attr:`folder_path` is
      ``'{defense_dir}/{dataset.data_type}/{dataset.name}/{model.name}/{defense.name}'``.
    | For defense implementation, see :class:`Defense`.

    Args:
        defense_name (str): The defense name.
        defense (str | Defense): The defense instance or defense name
            (as the alias of `defense_name`).
        dataset_name (str): The dataset name.
        dataset (str | trojanzoo.datasets.Dataset):
            Dataset instance or dataset name
            (as the alias of `dataset_name`).
        model_name (str): The model name.
        model (str | Model): The model instance or model name
            (as the alias of `model_name`).
        config (Config): The default parameter config.
        class_dict (dict[str, type[Defense]]):
            Map from defense name to defense class.
            Defaults to ``{}``.
        **kwargs: The keyword arguments
            passed to defense init method.

    Returns:
        Defense: The defense instance.
    """
    dataset_name = get_name(
        name=dataset_name, module=dataset, arg_list=['-d', '--dataset'])
    model_name = get_name(name=model_name, module=model,
                          arg_list=['-m', '--model'])
    defense_name = get_name(
        name=defense_name, module=defense, arg_list=['--defense'])
    if dataset_name is None:
        dataset_name = config.full_config['dataset']['default_dataset']
    general_config = config.get_config(dataset_name=dataset_name)['defense']
    specific_config = config.get_config(
        dataset_name=dataset_name)[defense_name]
    result = general_config.update(specific_config).update(
        kwargs)    # TODO: linting issues
    try:
        DefenseType = class_dict[defense_name]
    except KeyError:
        print(f'{defense_name} not in \n{list(class_dict.keys())}')
        raise
    if 'folder_path' not in result.keys():
        folder_path = result['defense_dir']
        if isinstance(dataset, Dataset):
            folder_path = os.path.join(
                folder_path, dataset.data_type, dataset.name)
        if model_name is not None:
            folder_path = os.path.join(folder_path, model_name)
        folder_path = os.path.join(folder_path, DefenseType.name)
        result['folder_path'] = folder_path
    return DefenseType(name=defense_name, dataset=dataset, model=model, **result)



#!/usr/bin/env python3

from . import operations

from trojanvision.utils.transform import cutout
# import torchvision.transforms.functional_tensor as F_t
# import torchvision.transforms.functional as F
from . import functional as F
from . import functional_tensor as F_t

import torch
import torch.nn as nn
import torch.nn.utils.parametrize as parametrize
from torch.autograd import Function


PRIMITIVES = ['Zero',
              'ShearX', 'ShearY',
              'TranslateX', 'TranslateY',
              'HorizontalFlip', 'Rotate',
              'Brightness', 'Contrast', 'AutoContrast',
              'Color', 'Sharpness',
              'Posterize', 'Solarize',
              'Equalize', 'Invert',
              'Cutout', 'SamplePairing']


def get_op(primitive: str, **kwargs) -> 'Operation':
    OpClass: type[Operation] = getattr(operations, primitive)
    return OpClass(**kwargs)


class Clamp(nn.Module):
    def __init__(self, _min: float = 0.0, _max: float = 1.0):
        super().__init__()
        self._min = _min
        self._max = _max
        self.scale = _max - _min

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        return X.clamp(0.0, 1.0) * self.scale + self._min


class _STE(Function):
    @staticmethod
    def forward(ctx,
                tensor: torch.Tensor,
                param: torch.Tensor):
        ctx.shape = param.shape
        return tensor.clone()

    @staticmethod
    def backward(ctx, grad_output: torch.Tensor):
        return grad_output.clone(), grad_output.sum_to_size(ctx.shape)


def ste(tensor: torch.Tensor, *params: torch.Tensor) -> torch.Tensor:
    if len(params) == 0:
        return tensor.clone()
    return ste(_STE.apply(tensor, params[0]), *params[1:])


class Operation(nn.Module):
    def __init__(self, **kwargs):
        super().__init__()

    def forward(self, _input: torch.Tensor) -> torch.Tensor:
        return _input

    def create_transform(self):
        return self


class MagnitudeOperation(Operation):
    def __init__(self, value_range: tuple[float, float] = (0.0, 1.0),
                 **kwargs):
        super().__init__()
        self.value_range = value_range
        self.magnitude = nn.Parameter(torch.rand(1))
        parametrize.register_parametrization(self, 'magnitude',
                                             Clamp(*value_range))

    def set_magnitude(self, magnitude: float):
        if parametrize.is_parametrized(self, 'magnitude'):
            mag = (magnitude - self.value_range[0]) / (self.value_range[1] - self.value_range[0])
            self.parametrizations.magnitude.original.data.fill_(mag)
        else:
            self.magnitude.data.fill_(magnitude)

    def create_transform(self):
        parametrize.remove_parametrizations(self, 'magnitude')
        self.requires_grad_(False)
        self.cpu()
        self.eval()
        return self


class Zero(Operation):
    def create_transform(self):
        return nn.Identity()


class ShearX(MagnitudeOperation):
    def __init__(self, value_range: tuple[float, float] = (-16.7, 16.7),
                 **kwargs):
        super().__init__(value_range=value_range, **kwargs)

    def forward(self, _input: torch.Tensor) -> torch.Tensor:
        shear = torch.cat([self.magnitude, torch.zeros_like(self.magnitude)])
        return F.affine(_input, shear=shear)


class ShearY(MagnitudeOperation):
    def __init__(self, value_range: tuple[float, float] = (-16.7, 16.7),
                 **kwargs):
        super().__init__(value_range=value_range, **kwargs)

    def forward(self, _input: torch.Tensor) -> torch.Tensor:
        shear = torch.cat([torch.zeros_like(self.magnitude), self.magnitude])
        return F.affine(_input, shear=shear)


class TranslateX(MagnitudeOperation):
    def __init__(self, value_range: tuple[float, float] = (-0.45, 0.45),
                 **kwargs):
        super().__init__(value_range=value_range, **kwargs)

    def forward(self, _input: torch.Tensor) -> torch.Tensor:
        translate = torch.cat([self.magnitude * _input.shape[-1],
                               torch.zeros_like(self.magnitude)])
        return F.affine(_input, translate=translate)


class TranslateY(MagnitudeOperation):
    def __init__(self, value_range: tuple[float, float] = (-0.45, 0.45),
                 **kwargs):
        super().__init__(value_range=value_range, **kwargs)

    def forward(self, _input: torch.Tensor) -> torch.Tensor:
        translate = torch.cat([torch.zeros_like(self.magnitude),
                               self.magnitude * _input.shape[-2]])
        return F.affine(_input, translate=translate)


class HorizontalFlip(Operation):
    def forward(self, _input: torch.Tensor) -> torch.Tensor:
        # return _input.flip(-1)
        return F_t.hflip(_input)


class Rotate(MagnitudeOperation):
    def __init__(self, value_range: tuple[float, float] = (-30.0, 30.0),
                 **kwargs):
        super().__init__(value_range=value_range, **kwargs)

    def forward(self, _input: torch.Tensor) -> torch.Tensor:
        # not differential w.r.t. self.magnitude
        # return F.rotate(_input, angle=self.magnitude)
        return F.affine(_input, angle=self.magnitude)

    #     rot = self.magnitude * torch.pi / 180    # radians
    #     matrix = self._get_matrix(rot)
    #     F_t.rotate(_input, matrix=matrix)

    # @staticmethod
    # def _get_matrix(rot: torch.Tensor) -> list[float]:
    #     cos, sin = rot.cos(), rot.sin()
    #     zero = torch.zeros_like(rot)
    #     matrix = torch.cat([cos, sin, zero, -sin, cos, zero])
    #     return matrix

    # @staticmethod
    # def rotate(img: torch.Tensor, matrix: torch.Tensor,
    #            interpolation: str = 'nearest',
    #            expand: bool = False, fill: None | list[float] = None
    #            ) -> torch.Tensor:
    #     w, h = img.shape[-1], img.shape[-2]
    #     ow, oh = F_t._compute_output_size(matrix, w, h) if expand else (w, h)
    #     theta = matrix.view(1, 2, 3)
    #     # grid will be generated on the same device as theta and img
    #     grid = F_t._gen_affine_grid(theta, w=w, h=h, ow=ow, oh=oh)
    #     return F_t._apply_grid_transform(img, grid, interpolation, fill=fill)


class Invert(Operation):
    def forward(self, _input: torch.Tensor) -> torch.Tensor:
        # return 1.0 - _input
        return F_t.invert(_input)


class Color(MagnitudeOperation):
    # alias of Saturate
    def forward(self, _input: torch.Tensor) -> torch.Tensor:
        # degenerate = F_t.rgb_to_grayscale(_input)
        # result: torch.Tensor = self.magnitude * _input \
        #     + (1 - self.magnitude) * degenerate
        # return result
        return F_t.adjust_saturation(_input, saturation_factor=self.magnitude)


class Brightness(MagnitudeOperation):
    def __init__(self, value_range: tuple[float, float] = (0.0, 2.0),
                 **kwargs):
        super().__init__(value_range=value_range, **kwargs)

    def forward(self, _input: torch.Tensor) -> torch.Tensor:
        # return (self.magnitude * _input).clamp(0.0, 1.0)
        return F_t.adjust_brightness(_input, brightness_factor=self.magnitude)


class Contrast(MagnitudeOperation):
    def __init__(self, value_range: tuple[float, float] = (0.0, 2.0),
                 **kwargs):
        super().__init__(value_range=value_range, **kwargs)

    def forward(self, _input: torch.Tensor) -> torch.Tensor:
        # degenerate = F_t.rgb_to_grayscale(_input).flatten(1).mean()
        # result: torch.Tensor = self.magnitude * _input \
        #     + (1 - self.magnitude) * degenerate
        # return result.clamp(0.0, 1.0)
        return F_t.adjust_contrast(_input, contrast_factor=self.magnitude)


class AutoContrast(Operation):
    def forward(self, _input: torch.Tensor) -> torch.Tensor:
        return F_t.autocontrast(_input)


class Sharpness(MagnitudeOperation):
    def __init__(self, value_range: tuple[float, float] = (0.0, 2.0),
                 **kwargs):
        super().__init__(value_range=value_range, **kwargs)

    def forward(self, _input: torch.Tensor) -> torch.Tensor:
        return F_t.adjust_sharpness(_input, sharpness_factor=self.magnitude)


class Equalize(Operation):
    def forward(self, _input: torch.Tensor) -> torch.Tensor:
        return ste(F_t.equalize((_input * 255).byte()).float().div(255), _input)


class Posterize(MagnitudeOperation):
    def __init__(self, value_range: tuple[float, float] = (0.0, 8.0),
                 **kwargs):
        super().__init__(value_range=value_range, ste=ste, **kwargs)

    def forward(self, _input: torch.Tensor) -> torch.Tensor:
        # not differential w.r.t. self.magnitude
        return ste(F_t.posterize((_input * 255).byte(), bits=int(self.magnitude)).float().div(255),
                   _input, self.magnitude)


class Solarize(MagnitudeOperation):
    """Invert all pixel values above a threshold."""

    def forward(self, _input: torch.Tensor) -> torch.Tensor:
        # not differential w.r.t. self.magnitude
        # return torch.where(_input > self.magnitude, 1 - _input, _input)
        return ste(F_t.solarize(_input, threshold=self.magnitude), _input, self.magnitude)


class Cutout(MagnitudeOperation):
    def __init__(self, value_range: tuple[float, float] = (0.0, 0.2),
                 **kwargs):
        super().__init__(value_range=value_range, **kwargs)

    def forward(self, _input: torch.Tensor) -> torch.Tensor:
        h, w = _input.shape[-2:]
        length = (self.magnitude * torch.tensor([h, w], dtype=torch.float, device=self.magnitude.device)).int()
        return ste(cutout(_input, length), _input, self.magnitude)


class SamplePairing(MagnitudeOperation):
    def __init__(self, value_range: tuple[float, float] = (0.0, 0.4),
                 **kwargs):
        super().__init__(value_range=value_range, **kwargs)

    def forward(self, _input: torch.Tensor) -> torch.Tensor:
        idx = torch.randperm(_input.shape[0])
        degenerate = _input[idx]
        result: torch.Tensor = (1 - self.magnitude) * _input \
            + self.magnitude * degenerate
        return result



#!/usr/bin/env python3

from ...abstract import CleanLabelBackdoor

from .clean_label import CleanLabel
from .invisible_poison import InvisiblePoison
from .refool import Refool


__all__ = ['CleanLabel', 'InvisiblePoison', 'Refool']

class_dict: dict[str, type[CleanLabelBackdoor]] = {
    'clean_label': CleanLabel,
    'invisible_poison': InvisiblePoison,
    'refool': Refool,
}



#!/usr/bin/env python3

import trojanvision
from trojanvision.attacks import PoisonRandom
import argparse

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    trojanvision.environ.add_argument(parser)
    trojanvision.datasets.add_argument(parser)
    trojanvision.models.add_argument(parser)
    trojanvision.trainer.add_argument(parser)
    trojanvision.attacks.add_argument(parser)
    kwargs = parser.parse_args().__dict__

    env = trojanvision.environ.create(**kwargs)
    dataset = trojanvision.datasets.create(**kwargs)
    model = trojanvision.models.create(dataset=dataset, **kwargs)
    trainer = trojanvision.trainer.create(dataset=dataset, model=model, **kwargs)
    attack: PoisonRandom = trojanvision.attacks.create(dataset=dataset, model=model, **kwargs)

    if env['verbose']:
        trojanvision.summary(env=env, dataset=dataset, model=model, train=trainer, attack=attack)
    attack.load()
    attack.validate_fn()



#!/usr/bin/env python3

import trojanvision
from trojanvision.utils import superimpose
import torchvision.transforms.functional as F
import argparse
import os

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    trojanvision.environ.add_argument(parser)
    trojanvision.datasets.add_argument(parser)
    trojanvision.models.add_argument(parser)
    kwargs = parser.parse_args().__dict__

    env = trojanvision.environ.create(**kwargs)
    dataset = trojanvision.datasets.create(**kwargs)
    model = trojanvision.models.create(dataset=dataset, **kwargs)

    if not os.path.exists('./result'):
        os.makedirs('./result')

    if env['verbose']:
        trojanvision.summary(env=env, dataset=dataset, model=model)
    for data in dataset.loader['valid']:
        _input, _label = model.get_data(data)
        heatmap = model.get_heatmap(_input, _label, method='saliency_map')
        heatmap = superimpose(heatmap, _input, alpha=0.5)
        for i, map in enumerate(heatmap):
            F.to_pil_image(heatmap[i]).save(f'./result/heatmap_{i}.jpg')
        break



#!/usr/bin/env python3

from ...abstract import TrainingFiltering
from .activation_clustering import ActivationClustering
# from .spectral_signature import SpectralSignature

__all__ = ['ActivationClustering']   # 'SpectralSignature',

class_dict: dict[str, type[TrainingFiltering]] = {
    'activation_clustering': ActivationClustering,
    # 'spectral_signature': SpectralSignature,
}



#!/usr/bin/env python3

r"""
CUDA_VISIBLE_DEVICES=0 python examples/backdoor_attack.py --color --verbose 1 --pretrained --validate_interval 1 --epochs 10 --lr 0.01 --attack input_aware_dynamic
"""  # noqa: E501

from ...abstract import BackdoorAttack

from trojanzoo.environ import env
from trojanzoo.utils.data import sample_batch
from trojanzoo.utils.logger import MetricLogger
from trojanzoo.utils.output import ansi, get_ansi_len, output_iter

import torch
import torch.nn as nn
from torchvision.models.resnet import conv3x3

import math
import random

import os

import argparse
from collections.abc import Callable


class InputAwareDynamic(BackdoorAttack):
    r"""Input-Aware Dynamic Backdoor Attack proposed by Anh Nguyen and Anh Tran
    from VinAI Research in NIPS 2020.

    Based on :class:`trojanvision.attacks.BadNet`,
    InputAwareDynamic trains mark generator and mask generator
    to synthesize unique watermark for each input.

    In classification loss, besides attacking poison inputs and classifying clean inputs,
    InputAwareDynamic also requires inputs attached
    with triggers generated from other inputs
    are still classified correctly (cross-trigger mode).

    See Also:
        * paper: `Input-Aware Dynamic Backdoor Attack`_
        * code: https://github.com/VinAIResearch/input-aware-backdoor-attack-release

    .. math::
       \begin{aligned}
            &\textbf{\# train mask generator}                                                                                                      \\
            &{opt}_{mask} = \text{Adam}(G_{mask}.parameters(), \text{lr}=0.01, \text{betas}=(0.5, 0.9))                                            \\
            &\textbf{for} \: e=1 \: \textbf{to} \: \text{train\_mask\_epochs}                                                                      \\
            &\hspace{5mm}\textbf{for} \: x_1 \: \textbf{in} \: \text{train\_set}                                                                   \\
            &\hspace{10mm}x_2 = \text{sample\_another\_batch}(\text{train\_set})                                                                   \\
            &\hspace{10mm}\mathcal{L}_{div}  = \frac{\lVert x_1 - x_2 \rVert}{\lVert G_{mask}(x_1) - G_{mask}(x_2) \rVert}                         \\
            &\hspace{10mm}\mathcal{L}_{norm} = ReLU(G_{mask}(x_1) - \text{mask\_density}).mean()                                                   \\
            &\hspace{10mm}\mathcal{L}_{mask} = \lambda_{div} \mathcal{L}_{div} + \lambda_{norm} \mathcal{L}_{norm}                                 \\
            &\hspace{10mm}{opt}_{mask}.step()                                                                                                      \\
            &\rule{110mm}{0.4pt}                                                                                                                   \\
            &\textbf{\# train mark generator and model}                                                                                            \\
            &{opt}_{mark} = \text{Adam}(G_{mark}.parameters(), \text{lr}=0.01, \text{betas}=(0.5, 0.9))                                            \\
            &\textbf{for} \: e=1 \: \textbf{to} \: \text{epochs}                                                                                   \\
            &\hspace{5mm}\textbf{for} \: (x_1, y_1) \: \textbf{in} \: \text{train\_set}                                                            \\
            &\hspace{10mm}x_2 = \text{sample\_another\_batch}(\text{train\_set})                                                                   \\
            &\hspace{10mm}{mark}_{poison}, {mask}_{poison} = G_{mark}, G_{mask} (x_1[:n_{poison}])                                                 \\
            &\hspace{10mm}{mark}_{cross}, {mask}_{cross}   = G_{mark}, G_{mask} (x_2[n_{poison}: n_{poison} + n_{cross}])                          \\
            &\hspace{10mm}x_{poison} = {mask}_{poison} \cdot {mark}_{poison} + (1 - {mask}_{poison}) \cdot x_1[:n_{poison}]                        \\
            &\hspace{10mm}x_{cross}  = {mask}_{cross}  \cdot {mark}_{cross}  + (1 - {mask}_{cross})  \cdot x_1[n_{poison}: n_{poison} + n_{cross}] \\
            &\hspace{10mm}x = cat([x_{poison}, x_{cross}, x_1[n_{poison}+n_{cross}:]])                                                             \\
            &\hspace{10mm}y = cat([y_{poison}, y_1[n_{poison}:]])                                                                                  \\
            &\hspace{10mm}\mathcal{L}_{div} = \frac{\lVert x_{poison} - x_{cross} \rVert}{\lVert {mark}_{poison} - {mark}_{cross} \rVert}          \\
            &\hspace{10mm}\mathcal{L}_{ce}  = cross\_entropy(x, y)                                                                                 \\
            &\hspace{10mm}\mathcal{L}       = \mathcal{L}_{ce} + \lambda_{div}\mathcal{L}_{div}                                                    \\
            &\hspace{10mm}{opt}_{mark}.step()                                                                                                      \\
            &\hspace{10mm}{opt}_{model}.step()                                                                                                     \\
       \end{aligned}

    Args:
        train_mask_epochs (int): Epoch to optimize mask generator.
            Defaults to ``25``.
        lambda_div (float): Weight of diversity loss
            during both optimization processes.
            Defaults to ``1.0``.
        lambda_norm (float): Weight of norm loss
            when optimizing mask generator.
            Defaults to ``100.0``.
        mask_density (float): Threshold of mask values
            when optimizing norm loss.
            Defaults to ``0.032``.
        cross_percent (float): Percentage of cross inputs
            in the whole training set.
            Defaults to ``0.1``.
        poison_percent (float): Percentage of poison inputs
            in the whole training set.
            Defaults to ``0.1``.
        natural (bool): Whether to use natural backdoors.
            If ``True``, model parameters will be frozen.
            Defaults to ``False``.

    Attributes:
        mark_generator (torch.nn.Sequential): Mark generator instance
            constructed by :meth:`define_generator()`.
            Output shape ``(N, C, H, W)``.
        mask_generator (torch.nn.Sequential): Mark generator instance
            constructed by :meth:`define_generator()`.
            Output shape ``(N, 1, H, W)``.

    Note:
        Do **NOT** directly call :attr:`self.mark_generator`
        or :attr:`self.mask_generator`.
        Their raw outputs are not normalized into range ``[0, 1]``.
        Please call :meth:`get_mark()` and :meth:`get_mask()` instead.


    .. _Input-Aware Dynamic Backdoor Attack:
        https://arxiv.org/abs/2010.08138
    """  # noqa: E501
    name: str = 'input_aware_dynamic'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--train_mask_epochs', type=int,
                           help='Epoch to optimize mask generator '
                           'before optimizing mark generator and model '
                           '(default: 25)')
        group.add_argument('--lambda_div', type=float,
                           help='weight of diversity loss '
                           'during both optimization processes '
                           '(default: 1.0)')
        group.add_argument('--lambda_norm', type=float,
                           help='weight of norm loss '
                           'when optimizing mask generator '
                           '(default: 100.0)')
        group.add_argument('--mask_density', type=float,
                           help='threshold of mask values '
                           'when optimizing norm loss '
                           '(default: 0.032)')
        group.add_argument('--cross_percent', type=float,
                           help='percentage of cross inputs '
                           'in the whole training set '
                           '(default: 0.032)')
        group.add_argument('--natural', action='store_true',
                           help='whether to use natural backdoors. '
                           'if true, model parameters will be frozen')
        return group

    def __init__(self, train_mask_epochs: int = 25,
                 lambda_div: float = 1.0, lambda_norm: float = 100.0,
                 mask_density: float = 0.032,
                 cross_percent: float = 0.1,
                 natural: bool = False,
                 poison_percent: float = 0.1, **kwargs):
        super().__init__(poison_percent=poison_percent, **kwargs)
        self.param_list['input_aware_dynamic'] = ['train_mask_epochs',
                                                  'lambda_div', 'lambda_norm',
                                                  'mask_density', 'cross_percent']

        self.train_mask_epochs = train_mask_epochs
        self.lambda_div = lambda_div
        self.lambda_norm = lambda_norm
        self.mask_density = mask_density
        self.natural = natural

        self.poison_ratio = self.poison_percent
        self.poison_num = 0  # monkey patch: to avoid batch size change in badnet
        self.cross_percent = cross_percent

        data_channel = self.dataset.data_shape[0]
        num_channels = [16, 32] if data_channel == 1 else [32, 64, 128]
        self.mark_generator = self.define_generator(
            num_channels, in_channels=data_channel
        ).to(device=env['device']).eval()
        self.mask_generator = self.define_generator(
            num_channels, in_channels=data_channel,
            out_channels=1).to(device=env['device']).eval()

        self.train_set = self.dataset.loader['train'].dataset
        self.idx = torch.randperm(len(self.train_set))
        self.pos = 0

    def add_mark(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
        r"""Add watermark to input tensor by calling
        :meth:`get_mark()` and :meth:`get_mask()`."""
        mark = self.get_mark(x)
        mask = self.get_mask(x)
        return x + mask * (mark - x)

    def get_mark(self, _input: torch.Tensor) -> torch.Tensor:
        r"""Get mark with shape ``(N, C, H, W)``.

        .. math::
            \begin{aligned}
                &raw = \text{self.mark\_generator(input)} \\
                &\textbf{return} \frac{\tanh{(raw)} + 1}{2}
            \end{aligned}
        """
        raw_output: torch.Tensor = self.mark_generator(_input)
        return raw_output.tanh() / 2 + 0.5

    def get_mask(self, _input: torch.Tensor) -> torch.Tensor:
        r"""Get mask with shape ``(N, 1, H, W)``.

        .. math::
            \begin{aligned}
                &raw = \text{self.mask\_generator(input)} \\
                &\textbf{return} \frac{\tanh{[10 \cdot \tanh{(raw)}]} + 1}{2}
            \end{aligned}
        """
        raw_output: torch.Tensor = self.mask_generator(_input)
        return raw_output.tanh().mul(10).tanh() / 2 + 0.5

    def get_data(self, data: tuple[torch.Tensor, torch.Tensor],
                 org: bool = False, keep_org: bool = True,
                 poison_label: bool = True, **kwargs
                 ) -> tuple[torch.Tensor, torch.Tensor]:
        r"""Get data.

        Note:
            The difference between this and
            :meth:`trojanvision.attacks.BadNet.get_data()` is:

            This method replaces some clean data with poison version,
            while BadNet's keeps the clean data and append poison version.
        """
        _input, _label = self.model.get_data(data)
        if not org:
            if keep_org:
                decimal, integer = math.modf(len(_label) * self.poison_percent)
                integer = int(integer)
                if random.uniform(0, 1) < decimal:
                    integer += 1
            else:
                integer = len(_label)
            if not keep_org or integer:
                trigger_input = self.add_mark(_input[:integer])
                _input = torch.cat([trigger_input, _input[integer:]])
                if poison_label:
                    trigger_label = self.target_class * torch.ones_like(_label[:integer])
                    _label = torch.cat([trigger_label, _label[integer:]])
        return _input, _label

    def _get_cross_data(self, data: tuple[torch.Tensor, torch.Tensor],
                        **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
        r"""Get cross-trigger mode data.
        Sample another batch from train set
        and apply their marks and masks to current batch.
        """
        _input, _label = self.model.get_data(data)
        batch_size = len(_input)
        data2 = sample_batch(self.train_set, idx=self.idx[self.pos:self.pos + batch_size])
        _input2, _label2 = self.model.get_data(data2)
        self.pos += batch_size
        if self.pos >= len(self.idx):
            self.pos = 0
            self.idx = torch.randperm(len(self.idx))
        mark, mask = self.get_mark(_input2), self.get_mask(_input2)
        _input = _input + mask * (mark - _input)
        return _input, _label

    def validate_fn(self,
                    get_data_fn: Callable[..., tuple[torch.Tensor, torch.Tensor]] = None,
                    loss_fn: Callable[..., torch.Tensor] = None,
                    main_tag: str = 'valid', indent: int = 0,
                    threshold: float = 5.0,
                    **kwargs) -> tuple[float, float]:
        clean_acc, _ = self.model._validate(print_prefix='Validate Clean', main_tag='valid clean',
                                            get_data_fn=None, indent=indent, **kwargs)
        asr, _ = self.model._validate(print_prefix='Validate ASR', main_tag='valid asr',
                                      get_data_fn=self.get_data, keep_org=False, poison_label=True,
                                      indent=indent, **kwargs)
        self.model._validate(print_prefix='Validate Cross', main_tag='valid cross',
                             get_data_fn=self._get_cross_data, indent=indent, **kwargs)
        # prints(f'Validate Confidence: {self.validate_confidence():.3f}', indent=indent)
        # prints(f'Neuron Jaccard Idx: {self.get_neuron_jaccard():.3f}', indent=indent)
        if self.clean_acc - clean_acc > threshold:
            asr = 0.0
        return asr, clean_acc

    def attack(self, epochs: int, optimizer: torch.optim.Optimizer,
               lr_scheduler: torch.optim.lr_scheduler._LRScheduler = None,
               validate_interval: int = 1, save: bool = False,
               verbose: bool = True, **kwargs):
        if verbose:
            print('train mask generator')
        self.mark_generator.requires_grad_(False)
        self.mask_generator.requires_grad_()
        self.model.requires_grad_(False)
        self.train_mask_generator(verbose=verbose)
        if verbose:
            print()
            print('train mark generator and model')

        self.mark_generator.requires_grad_()
        self.mask_generator.requires_grad_(False)
        if not self.natural:
            params: list[nn.Parameter] = []
            for param_group in optimizer.param_groups:
                params.extend(param_group['params'])
            self.model.activate_params(params)

        mark_optimizer = torch.optim.Adam(self.mark_generator.parameters(), lr=1e-2, betas=(0.5, 0.9))
        mark_scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
            mark_optimizer, T_max=epochs)
        loader = self.dataset.loader['train']
        dataset = loader.dataset
        logger = MetricLogger()
        logger.create_meters(loss=None, div=None, ce=None)

        if validate_interval != 0:
            best_validate_result = self.validate_fn(verbose=verbose)
            best_asr = best_validate_result[0]
        for _epoch in range(epochs):
            _epoch += 1
            idx = torch.randperm(len(dataset))
            pos = 0
            logger.reset()
            if not self.natural:
                self.model.train()
            self.mark_generator.train()
            header: str = '{blue_light}{0}: {1}{reset}'.format(
                'Epoch', output_iter(_epoch, epochs), **ansi)
            header = header.ljust(max(len('Epoch'), 30) + get_ansi_len(header))
            for data in logger.log_every(loader, header=header) if verbose else loader:
                if not self.natural:
                    optimizer.zero_grad()
                mark_optimizer.zero_grad()
                _input, _label = self.model.get_data(data)
                batch_size = len(_input)
                data2 = sample_batch(dataset, idx=idx[pos:pos + batch_size])
                _input2, _label2 = self.model.get_data(data2)
                pos += batch_size
                final_input, final_label = _input.clone(), _label.clone()

                # generate trigger input
                trigger_dec, trigger_int = math.modf(len(_label) * self.poison_percent)
                trigger_int = int(trigger_int)
                if random.uniform(0, 1) < trigger_dec:
                    trigger_int += 1
                x = _input[:trigger_int]
                trigger_mark, trigger_mask = self.get_mark(x), self.get_mask(x)
                trigger_input = x + trigger_mask * (trigger_mark - x)
                final_input[:trigger_int] = trigger_input
                final_label[:trigger_int] = self.target_class

                # generate cross input
                cross_dec, cross_int = math.modf(len(_label) * self.cross_percent)
                cross_int = int(cross_int)
                if random.uniform(0, 1) < cross_dec:
                    cross_int += 1
                x = _input[trigger_int:trigger_int + cross_int]
                x2 = _input2[trigger_int:trigger_int + cross_int]
                cross_mark, cross_mask = self.get_mark(x2), self.get_mask(x2)
                cross_input = x + cross_mask * (cross_mark - x)
                final_input[trigger_int:trigger_int + cross_int] = cross_input

                # div loss
                if len(trigger_input) <= len(cross_input):
                    length = len(trigger_input)
                    cross_input = cross_input[:length]
                    cross_mark = cross_mark[:length]
                    cross_mask = cross_mask[:length]
                else:
                    length = len(cross_input)
                    trigger_input = trigger_input[:length]
                    trigger_mark = trigger_mark[:length]
                    trigger_mask = trigger_mask[:length]
                input_dist: torch.Tensor = (trigger_input - cross_input).flatten(1).norm(p=2, dim=1)
                mark_dist: torch.Tensor = (trigger_mark - cross_mark).flatten(1).norm(p=2, dim=1) + 1e-5

                loss_ce = self.model.loss(final_input, final_label)
                loss_div = input_dist.div(mark_dist).mean()

                loss = loss_ce + self.lambda_div * loss_div
                loss.backward()
                if not self.natural:
                    optimizer.step()
                mark_optimizer.step()
                logger.update(n=batch_size, loss=loss.item(), div=loss_div.item(), ce=loss_ce.item())
            if not self.natural and lr_scheduler:
                lr_scheduler.step()
            mark_scheduler.step()
            if not self.natural:
                self.model.eval()
            self.mark_generator.eval()
            if validate_interval != 0 and (_epoch % validate_interval == 0 or _epoch == epochs):
                validate_result = self.validate_fn(verbose=verbose)
                cur_asr = validate_result[0]
                if cur_asr >= best_asr:
                    best_validate_result = validate_result
                    best_asr = cur_asr
                    if save:
                        self.save()
        if not self.natural:
            optimizer.zero_grad()
        mark_optimizer.zero_grad()
        self.mark_generator.requires_grad_(False)
        self.mask_generator.requires_grad_(False)
        self.model.requires_grad_(False)
        return best_validate_result

    def save(self, filename: str = None, **kwargs):
        r"""Save attack results to files."""
        filename = filename or self.get_filename(**kwargs)
        file_path = os.path.join(self.folder_path, filename)
        torch.save(self.mask_generator.state_dict(), file_path + '_mask.pth')
        torch.save(self.mark_generator.state_dict(), file_path + '_mark.pth')
        self.model.save(file_path + '.pth')
        print('attack results saved at: ', file_path)

    def load(self, filename: str = None, **kwargs):
        r"""Load attack results from previously saved files."""
        filename = filename or self.get_filename(**kwargs)
        file_path = os.path.join(self.folder_path, filename)
        self.mask_generator.load_state_dict(torch.load(file_path + '_mask.pth'))
        self.mark_generator.load_state_dict(torch.load(file_path + '_mark.pth'))
        self.model.load(file_path + '.pth')
        print('attack results loaded from: ', file_path)

    def train_mask_generator(self, verbose: bool = True):
        r"""Train :attr:`self.mask_generator`."""
        optimizer = torch.optim.Adam(self.mask_generator.parameters(), lr=1e-2, betas=(0.5, 0.9))
        lr_scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
            optimizer, T_max=self.train_mask_epochs)
        loader = self.dataset.loader['train']
        dataset = loader.dataset
        logger = MetricLogger()
        logger.create_meters(loss=None, div=None, norm=None)
        print_prefix = 'Mask Epoch'
        for _epoch in range(self.train_mask_epochs):
            _epoch += 1
            idx = torch.randperm(len(dataset))
            pos = 0
            logger.reset()
            header: str = '{blue_light}{0}: {1}{reset}'.format(
                print_prefix, output_iter(_epoch, self.train_mask_epochs), **ansi)
            header = header.ljust(max(len(print_prefix), 30) + get_ansi_len(header))
            self.mask_generator.train()
            for data in logger.log_every(loader, header=header) if verbose else loader:
                optimizer.zero_grad()
                _input, _label = self.model.get_data(data)
                batch_size = len(_input)
                data2 = sample_batch(dataset, idx=idx[pos:pos + batch_size])
                _input2, _label2 = self.model.get_data(data2)
                pos += batch_size

                _mask = self.get_mask(_input)
                _mask2 = self.get_mask(_input2)

                input_dist: torch.Tensor = (_input - _input2).flatten(1).norm(p=2, dim=1)
                mask_dist: torch.Tensor = (_mask - _mask2).flatten(1).norm(p=2, dim=1) + 1e-5

                loss_div = input_dist.div(mask_dist).mean()
                loss_norm = _mask.sub(self.mask_density).relu().mean()

                loss = self.lambda_norm * loss_norm + self.lambda_div * loss_div
                loss.backward()
                optimizer.step()
                logger.update(n=batch_size, loss=loss.item(), div=loss_div.item(), norm=loss_norm.item())
            lr_scheduler.step()
            self.mask_generator.eval()
            if verbose and (_epoch % (max(self.train_mask_epochs // 5, 1)) == 0 or _epoch == self.train_mask_epochs):
                self.validate_mask_generator()
        optimizer.zero_grad()

    @torch.no_grad()
    def validate_mask_generator(self):
        loader = self.dataset.loader['valid']
        dataset = loader.dataset
        logger = MetricLogger()
        logger.create_meters(loss=None, div=None, norm=None)
        idx = torch.randperm(len(dataset))
        pos = 0

        print_prefix = 'Validate'
        header: str = '{yellow}{0}{reset}'.format(print_prefix, **ansi)
        header = header.ljust(max(len(print_prefix), 30) + get_ansi_len(header))
        for data in logger.log_every(loader, header=header):
            _input, _label = self.model.get_data(data)
            batch_size = len(_input)
            data2 = sample_batch(dataset, idx=idx[pos:pos + batch_size])
            _input2, _label2 = self.model.get_data(data2)
            pos += batch_size

            _mask = self.get_mask(_input)
            _mask2 = self.get_mask(_input2)

            input_dist: torch.Tensor = (_input - _input2).flatten(1).norm(p=2, dim=1)
            mask_dist: torch.Tensor = (_mask - _mask2).flatten(1).norm(p=2, dim=1) + 1e-5

            loss_div = input_dist.div(mask_dist).mean()
            loss_norm = _mask.sub(self.mask_density).relu().mean()

            loss = self.lambda_norm * loss_norm + self.lambda_div * loss_div
            logger.update(n=batch_size, loss=loss.item(), div=loss_div.item(), norm=loss_norm.item())

    @staticmethod
    def define_generator(num_channels: list[int] = [32, 64, 128],
                         in_channels: int = 3, out_channels: int = None
                         ) -> nn.Sequential:
        r"""Define a generator used in :attr:`self.mark_generator` and :attr:`self.mask_generator`.

        Similar to auto-encoders, the generator is composed of ``['down', 'middle', 'up']``.

        * **down**: :math:`[\text{conv-bn-relu}(c_{i}, c_{i+1}), \text{conv-bn-relu}(c_{i+1}, c_{i+1}), \text{maxpool}(2)]`
        * **middle**: :math:`[\text{conv-bn-relu}(c_{-1}, c_{-1})]`
        * **up**: :math:`[\text{upsample}(2), \text{conv-bn-relu}(c_{i+1}, c_{i+1}), \text{conv-bn-relu}(c_{i+1}, c_{i})]`

        Args:
            num_channels (list[int]): List of intermediate feature numbers.
                Each element serves as the :attr:`in_channels` of current layer
                and :attr:`out_features` of preceding layer.
                Defaults to ``[32, 64, 128]``.

                * MNIST: ``[16, 32]``
                * CIFAR: ``[32, 64, 128]``

            in_channels (int): :attr:`in_channels` of first conv layer in ``down``.
                It should be image channels.
                Defaults to ``3``.
            out_channels (int): :attr:`out_channels` of last conv layer in ``up``.
                Defaults to ``None`` (:attr:`in_channels`).

        Returns:
            torch.nn.Sequential: Generator instance with input shape ``(N, in_channels, H, W)``
                and output shape ``(N, out_channels, H, W)``.
        """  # noqa: E501
        out_channels = out_channels or in_channels
        down_channel_list = num_channels.copy()
        down_channel_list.insert(0, in_channels)
        up_channel_list = num_channels[::-1].copy()
        up_channel_list.append(out_channels)

        seq = nn.Sequential()
        down_seq = nn.Sequential()
        middle_seq = nn.Sequential()
        up_seq = nn.Sequential()
        for i in range(len(num_channels)):
            down_seq.add_module(f'conv{3*i+1}', conv3x3(down_channel_list[i], down_channel_list[i + 1]))
            down_seq.add_module(f'bn{3*i+1}', nn.BatchNorm2d(down_channel_list[i + 1], momentum=0.05))
            down_seq.add_module(f'relu{3*i+1}', nn.ReLU(inplace=True))
            down_seq.add_module(f'conv{3*i+2}', conv3x3(down_channel_list[i + 1], down_channel_list[i + 1]))
            down_seq.add_module(f'bn{3*i+2}', nn.BatchNorm2d(down_channel_list[i + 1], momentum=0.05))
            down_seq.add_module(f'relu{3*i+2}', nn.ReLU(inplace=True))
            down_seq.add_module(f'maxpool{3*i+3}', nn.MaxPool2d(kernel_size=2))
        middle_seq.add_module('conv', conv3x3(num_channels[-1], num_channels[-1]))
        middle_seq.add_module('bn', nn.BatchNorm2d(num_channels[-1], momentum=0.05))
        middle_seq.add_module('relu', nn.ReLU(inplace=True))
        for i in range(len(num_channels)):
            up_seq.add_module(f'upsample{3*i+1}', nn.Upsample(scale_factor=2.0))
            up_seq.add_module(f'conv{3*i+2}', conv3x3(up_channel_list[i], up_channel_list[i]))
            up_seq.add_module(f'bn{3*i+2}', nn.BatchNorm2d(up_channel_list[i], momentum=0.05))
            up_seq.add_module(f'relu{3*i+2}', nn.ReLU(inplace=True))
            up_seq.add_module(f'conv{3*i+3}', conv3x3(up_channel_list[i], up_channel_list[i + 1]))
            up_seq.add_module(f'bn{3*i+3}', nn.BatchNorm2d(up_channel_list[i + 1], momentum=0.05))
            up_seq.add_module(f'relu{3*i+3}', nn.ReLU(inplace=True))
        seq.add_module('down', down_seq)
        seq.add_module('middle', middle_seq)
        seq.add_module('up', up_seq)
        return seq



#!/usr/bin/env python3

from ...abstract import ModelInspection
from trojanvision.environ import env
from trojanzoo.utils.logger import AverageMeter
from trojanzoo.utils.output import prints, ansi, output_iter

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import time
import datetime
import argparse
from tqdm import tqdm


class DeepInspect(ModelInspection):

    name: str = 'deep_inspect'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--sample_ratio', type=float, help='sample ratio from the full training data')
        group.add_argument('--noise_dim', type=int, help='GAN noise dimension')
        group.add_argument('--gamma_1', type=float, help='control effect of GAN loss')
        group.add_argument('--gamma_2', type=float, help='control effect of perturbation loss')
        return group

    def __init__(self, defense_remask_epoch: int = 20, defense_remask_lr=0.01,
                 sample_ratio: float = 0.1, noise_dim: int = 100,
                 gamma_1: float = 0.0, gamma_2: float = 0.02, **kwargs):
        super().__init__(defense_remask_epoch=defense_remask_epoch, defense_remask_lr=defense_remask_lr, **kwargs)
        self.param_list['deep_inspect'] = ['sample_ratio', 'gamma_1', 'gamma_2']

        self.sample_ratio: int = sample_ratio
        self.defense_remask_lr: float = defense_remask_lr
        self.gamma_1: float = gamma_1
        self.gamma_2: float = gamma_2

        self.noise_dim: int = noise_dim

        dataset = self.dataset.get_dataset(mode='train')
        subset, _ = self.dataset.split_dataset(dataset, percent=sample_ratio)
        self.loader = self.dataset.get_dataloader(mode='train', dataset=subset)

    def optimize_mark(self, label: int, **kwargs) -> tuple[torch.Tensor, float]:
        r"""
        Args:
            label (int): The class label to optimize.
            **kwargs: Any keyword argument (unused).

        Returns:
            (torch.Tensor, torch.Tensor):
                Optimized mark tensor with shape ``(C + 1, H, W)``
                and loss tensor.
        """
        epochs = self.defense_remask_epoch
        generator = Generator(self.noise_dim, self.dataset.num_classes, self.dataset.data_shape)
        generator.requires_grad_()
        optimizer = optim.Adam(generator.parameters(), lr=self.defense_remask_lr)
        optimizer.zero_grad()

        losses = AverageMeter('Loss', ':.4e')
        entropy = AverageMeter('Entropy', ':.4e')
        norm = AverageMeter('Norm', ':.4e')
        acc = AverageMeter('Acc', ':6.2f')

        noise = torch.rand(1, self.noise_dim, device=env['device'])
        mark = torch.zeros(self.dataset.data_shape, device=env['device'])
        for _epoch in range(epochs):
            losses.reset()
            entropy.reset()
            norm.reset()
            acc.reset()
            epoch_start = time.perf_counter()
            loader = self.loader
            if env['tqdm']:
                loader = tqdm(loader, leave=False)
            for data in loader:
                _input, _label = self.model.get_data(data)
                mark: torch.Tensor = generator(noise, torch.tensor([label], device=_label.device, dtype=_label.dtype))
                self.attack.mark.mark = torch.ones_like(self.attack.mark.mark)
                self.attack.mark.mark[:-1] = mark.squeeze()
                # Or directly add and clamp according to their paper?
                trigger_input = self.attack.add_mark(_input)
                trigger_label = label * torch.ones_like(_label)
                trigger_output = self.model(trigger_input)

                batch_acc = trigger_label.eq(trigger_output.argmax(1)).float().mean()
                batch_entropy = self.loss(_input, _label,
                                          target=label,
                                          trigger_output=trigger_output)
                batch_norm = torch.mean(self.attack.mark.mark[:-1].norm(p=1))
                batch_loss = batch_entropy + self.gamma_2 * batch_norm

                batch_size = _label.size(0)
                acc.update(batch_acc.item(), batch_size)
                entropy.update(batch_entropy.item(), batch_size)
                norm.update(batch_norm.item(), batch_size)
                losses.update(batch_loss.item(), batch_size)

                batch_loss.backward()
                optimizer.step()
                optimizer.zero_grad()

            epoch_time = str(datetime.timedelta(seconds=int(
                time.perf_counter() - epoch_start)))
            pre_str: str = '{blue_light}Epoch: {0}{reset}'.format(
                output_iter(_epoch + 1, epochs), **ansi)
            pre_str = pre_str.ljust(64 if env['color'] else 35)
            _str = ' '.join([
                f'Loss: {losses.avg:.4f},'.ljust(20),
                f'Acc: {acc.avg:.2f}, '.ljust(20),
                f'Norm: {norm.avg:.4f},'.ljust(20),
                f'Entropy: {entropy.avg:.4f},'.ljust(20),
                f'Time: {epoch_time},'.ljust(20),
            ])
            prints(pre_str, _str, indent=4)
        generator.requires_grad_(False)
        return self.attack.mark.mark, losses.avg


class Generator(nn.Module):
    def __init__(self, noise_dim: int = 100, num_classes: int = 10, data_shape: list[int] = [3, 32, 32]):
        self.noise_dim: int = noise_dim
        self.num_classes: int = num_classes
        self.data_shape: list[int] = data_shape
        super().__init__()

        self.relu = nn.ReLU(inplace=True)
        self.sigmoid = nn.Sigmoid()
        self.fc2 = nn.Linear(self.num_classes, 1000)
        self.fc = nn.Linear(self.noise_dim + 1000, 64 * self.data_shape[1] * self.data_shape[2])
        self.bn1 = nn.BatchNorm2d(64)
        self.deconv1 = nn.ConvTranspose2d(64, 32, 5, 1, 2)
        self.bn2 = nn.BatchNorm2d(32)
        self.deconv2 = nn.ConvTranspose2d(32, self.data_shape[0], 5, 1, 2)
        if env['num_gpus']:
            self.cuda()

    def forward(self, noise: torch.Tensor, trigger_label: torch.Tensor) -> torch.Tensor:
        _label: torch.Tensor = F.one_hot(trigger_label, self.num_classes)
        _label = _label.float()
        y_ = self.fc2(_label)
        y_ = self.relu(y_)
        x = torch.cat([noise, y_], dim=1)
        x: torch.Tensor = self.fc(x)
        x = x.view(-1, 64, self.data_shape[1], self.data_shape[2])
        x = self.bn1(x)
        x = self.relu(x)
        x = self.deconv1(x)
        x = self.bn2(x)
        x = self.relu(x)
        x = self.deconv2(x)
        x = self.sigmoid(x)
        return x



#!/usr/bin/env python3

from trojanvision.datasets.imageset import ImageSet

import torchvision.datasets as datasets


class CIFAR10(ImageSet):
    r"""CIFAR10 dataset introduced by Alex Krizhevsky in 2009.
    It inherits :class:`trojanvision.datasets.ImageSet`.

    See Also:
        * torchvision: :any:`torchvision.datasets.CIFAR10`
        * paper: `Learning Multiple Layers of Features from Tiny Images`_
        * website: https://www.cs.toronto.edu/~kriz/cifar.html

    Attributes:
        name (str): ``'cifar10'``
        num_classes (int): ``10``
        data_shape (list[int]): ``[3, 32, 32]``
        class_names (list[str]):
            | ``['airplane', 'automobile', 'bird', 'cat', 'deer',``
            | ``'dog', 'frog', 'horse', 'ship', 'truck']``
        norm_par (dict[str, list[float]]):
            | ``{'mean': [0.49139968, 0.48215827, 0.44653124],``
            | ``'std'  : [0.24703233, 0.24348505, 0.26158768]}``

    .. _Learning Multiple Layers of Features from Tiny Images:
        https://www.cs.toronto.edu/~kriz/learning-features-2009-TR.pdf
    """
    name = 'cifar10'
    num_classes = 10
    data_shape = [3, 32, 32]
    class_names = ['airplane', 'automobile', 'bird', 'cat', 'deer',
                   'dog', 'frog', 'horse', 'ship', 'truck']

    def __init__(self, norm_par: dict[str, list[float]] = {'mean': [0.49139968, 0.48215827, 0.44653124],
                                                           'std': [0.24703233, 0.24348505, 0.26158768], },
                 **kwargs):
        super().__init__(norm_par=norm_par, **kwargs)

    def initialize(self):
        datasets.CIFAR10(root=self.folder_path, train=True, download=True)
        datasets.CIFAR10(root=self.folder_path, train=False, download=True)

    def _get_org_dataset(self, mode: str, **kwargs) -> datasets.CIFAR10:
        assert mode in ['train', 'valid']
        return datasets.CIFAR10(root=self.folder_path, train=(mode == 'train'), **kwargs)


class CIFAR100(CIFAR10):
    r"""CIFAR100 dataset. It inherits :class:`trojanvision.datasets.ImageSet`.

    See Also:
        * torchvision: :any:`torchvision.datasets.CIFAR100`
        * paper: `Learning Multiple Layers of Features from Tiny Images`_
        * website: https://www.cs.toronto.edu/~kriz/cifar.html

    Attributes:
        name (str): ``'cifar100'``
        num_classes (int): ``100``
        data_shape (list[int]): ``[3, 32, 32]``
        norm_par (dict[str, list[float]]):
            | ``{'mean': [0.49139968, 0.48215827, 0.44653124],``
            | ``'std'  : [0.24703233, 0.24348505, 0.26158768]}``

    .. _Learning Multiple Layers of Features from Tiny Images:
        https://www.cs.toronto.edu/~kriz/learning-features-2009-TR.pdf
    """
    name = 'cifar100'
    num_classes = 100

    def initialize(self):
        datasets.CIFAR100(root=self.folder_path, train=True, download=True)
        datasets.CIFAR100(root=self.folder_path, train=False, download=True)

    def _get_org_dataset(self, mode: str, **kwargs) -> datasets.CIFAR100:
        assert mode in ['train', 'valid']
        return datasets.CIFAR100(root=self.folder_path, train=(mode == 'train'), **kwargs)



#!/usr/bin/env python3

# CUDA_VISIBLE_DEVICES=0 python ./examples/backdoor_defense.py --color --verbose 1 --attack badnet --defense neural_cleanse --validate_interval 1 --epochs 50 --lr 1e-2

import trojanvision
import argparse

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    trojanvision.environ.add_argument(parser)
    trojanvision.datasets.add_argument(parser)
    trojanvision.models.add_argument(parser)
    trojanvision.trainer.add_argument(parser)
    trojanvision.marks.add_argument(parser)
    trojanvision.attacks.add_argument(parser)
    trojanvision.defenses.add_argument(parser)
    kwargs = parser.parse_args().__dict__

    env = trojanvision.environ.create(**kwargs)
    dataset = trojanvision.datasets.create(**kwargs)
    model = trojanvision.models.create(dataset=dataset, **kwargs)
    trainer = trojanvision.trainer.create(dataset=dataset, model=model, **kwargs)
    mark = trojanvision.marks.create(dataset=dataset, **kwargs)
    attack = trojanvision.attacks.create(dataset=dataset, model=model, mark=mark, **kwargs)
    defense = trojanvision.defenses.create(dataset=dataset, model=model, attack=attack, **kwargs)

    if env['verbose']:
        trojanvision.summary(env=env, dataset=dataset, model=model, mark=mark, trainer=trainer, attack=attack, defense=defense)
    defense.detect(**trainer)



#!/usr/bin/env python3

import torch


def normalize_mad(values: torch.Tensor | list[float], side: str = None) -> torch.Tensor:
    if not isinstance(values, torch.Tensor):
        values = torch.as_tensor(values, dtype=torch.float)
    median = values.median()
    abs_dev = (values - median).abs()
    mad = abs_dev.median()
    measures = abs_dev / (mad + 1e-8) / 1.4826
    if side == 'double':    # TODO: use a loop to optimize code
        dev_list = []
        for i in range(len(values)):
            if values[i] <= median:
                dev_list.append(float(median - values[i]))
        mad = torch.tensor(dev_list).median()
        for i in range(len(values)):
            if values[i] <= median:
                measures[i] = abs_dev[i] / mad / 1.4826

        dev_list = []
        for i in range(len(values)):
            if values[i] >= median:
                dev_list.append(float(values[i] - median))
        mad = torch.tensor(dev_list).median()
        for i in range(len(values)):
            if values[i] >= median:
                measures[i] = abs_dev[i] / mad / 1.4826
    return measures


def mask_jaccard(mask: torch.Tensor, real_mask: torch.Tensor,
                 select_num: int = 9) -> float:
    mask = mask.float()
    real_mask = real_mask.float()
    detect_mask = mask > mask.flatten().topk(select_num)[0][-1]
    sum_temp = detect_mask.int() + real_mask.int()
    overlap = (sum_temp == 2).sum().float() / (sum_temp >= 1).sum().float()
    return float(overlap)



#!/usr/bin/env python3

from trojanvision.optim import PGDoptimizer
from trojanzoo.attacks import Attack
from trojanzoo.utils.output import prints, ansi
from trojanzoo.utils.logger import SmoothedValue

import torch
import argparse
from collections.abc import Callable
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from trojanvision.datasets import ImageSet
    from trojanvision.models import ImageModel


def get_func_key(func: Callable[..., torch.Tensor]) -> str:
    keys = func.__code__.co_varnames
    valid_keys = [key for key in keys if 'target' in key or 'label' in key or 'y' in key or 'Y' in key]
    assert len(valid_keys) == 1, keys
    return valid_keys[0]


class PGD(Attack, PGDoptimizer):
    r"""PGD Adversarial Attack.

    Args:
        pgd_alpha (float): learning rate :math:`\alpha`. Default: :math:`\frac{3}{255}`.
        pgd_eps (float): the perturbation threshold :math:`\epsilon` in input space. Default: :math:`\frac{8}{255}`.
    """

    name: str = 'pgd'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--pgd_alpha', type=float, help='PGD learning rate per step, defaults to 2.0/255')
        group.add_argument('--pgd_eps', type=float, help='Projection norm constraint, defaults to 8.0/255')
        group.add_argument('--iteration', type=int, help='Attack Iteration, defaults to 7')
        group.add_argument('--stop_threshold', type=float, help='early stop confidence, defaults to 0.99')
        group.add_argument('--target_class', type=int, help='Do not set it if using target_idx')
        group.add_argument('--target_idx', type=int,
                           help='Target label order in original classification, defaults to -1 '
                           '(0 for untargeted attack, 1 for most possible class, -1 for most unpossible class)')
        group.add_argument('--test_num', type=int, help='total number of test examples for PGD, defaults to 1000.')
        group.add_argument('--num_restart', type=int,
                           help='number of random init for PGD, defaults to 0 (without random initialization).')
        group.add_argument('--require_class', action='store_true',
                           help='Require class to be desired in early_stop_check.')

        group.add_argument('--grad_method', choices=['white', 'nes', 'sgd', 'hess', 'zoo'],
                           help='gradient estimation method, defaults to "white"')
        group.add_argument('--query_num', type=int,
                           help='query numbers for black box gradient estimation, defaults to 100.')
        group.add_argument('--sigma', type=float,
                           help='gaussian sampling std for black box gradient estimation, defaults to 1e-3')
        return group

    def __init__(self, target_class: int = None, target_idx: int = -1, test_num: int = 1000, num_restart: int = 0,
                 require_class: bool = False, **kwargs):
        self.target_class = target_class
        self.target_idx = target_idx
        self.test_num = test_num
        self.num_restart = num_restart
        self.require_class = require_class
        kwargs.update(random_init=bool(num_restart))
        super().__init__(**kwargs)
        self.param_list['pgd_attack'] = ['target_class', 'target_idx', 'test_num', 'num_restart', 'require_class']
        self.dataset: ImageSet
        self.model: 'ImageModel'

    def attack(self, verbose: int = 1, **kwargs) -> tuple[float, float]:
        validset = self.dataset.get_dataset('valid')
        testset, _ = self.dataset.split_dataset(validset, percent=0.3)
        loader = self.dataset.get_dataloader(mode='valid', dataset=testset,
                                             shuffle=True)
        fmt_str = '{global_avg:7.3f} ({min:7.3f}  {max:7.3f})'
        total_adv_target_conf = SmoothedValue(fmt=fmt_str)
        total_org_target_conf = SmoothedValue(fmt=fmt_str)
        succ_adv_target_conf = SmoothedValue(fmt=fmt_str)

        total_adv_org_conf = SmoothedValue(fmt=fmt_str)
        total_org_org_conf = SmoothedValue(fmt=fmt_str)
        succ_adv_org_conf = SmoothedValue(fmt=fmt_str)

        total_iter_list = SmoothedValue(fmt=fmt_str)
        succ_iter_list = SmoothedValue(fmt=fmt_str)

        succ_idx_list: list[int] = []
        for data in loader:
            rest_length = self.test_num - total_adv_target_conf.count
            if rest_length <= 0:
                break
            _input, _label = self.model.remove_misclassify(data)
            if len(_label) == 0:
                continue

            if len(_label) > rest_length:
                _input = _input[:rest_length]
                _label = _label[:rest_length]
            target = self.generate_target(_input, idx=self.target_idx) if self.target_class is None \
                else self.target_class * torch.ones_like(_label)
            adv_input = _input.clone().detach()
            iter_list = -torch.ones(len(_label), dtype=torch.long)
            current_idx = torch.arange(len(iter_list))
            for _ in range(max(self.num_restart, 1)):
                temp_adv_input, temp_iter_list = self.optimize(_input[current_idx],
                                                               target=target[current_idx], **kwargs)
                adv_input[current_idx] = temp_adv_input
                iter_list[current_idx] = temp_iter_list
                fail_idx = iter_list == -1
                if (~fail_idx).all():
                    break
                current_idx = current_idx[fail_idx]
            for i, _iter in enumerate(iter_list):
                if _iter != -1:
                    succ_idx_list.append(total_iter_list.count + i)
            adv_target_conf = self.model.get_target_prob(adv_input, target)
            adv_org_conf = self.model.get_target_prob(adv_input, _label)
            org_target_conf = self.model.get_target_prob(_input, target)
            org_org_conf = self.model.get_target_prob(_input, _label)

            total_adv_target_conf.update_list(adv_target_conf.detach().cpu().tolist())
            total_adv_org_conf.update_list(adv_org_conf.detach().cpu().tolist())
            succ_adv_target_conf.update_list(adv_target_conf[iter_list != -1].detach().cpu().tolist())
            succ_adv_org_conf.update_list(adv_org_conf[iter_list != -1].detach().cpu().tolist())
            total_org_target_conf.update_list(org_target_conf.detach().cpu().tolist())
            total_org_org_conf.update_list(org_org_conf.detach().cpu().tolist())

            total_iter_list.update_list(torch.where(iter_list != -1, iter_list, 2 *
                                        self.iteration * torch.ones_like(iter_list)).tolist())
            succ_iter_list.update_list(iter_list[iter_list != -1].tolist())
            if verbose >= 3:
                prints(f'{ansi["green"]}{succ_iter_list.count} / {total_iter_list.count}{ansi["reset"]}')
            if verbose >= 4:
                prints(f'{total_iter_list=:}', indent=4)
                prints(f'{succ_iter_list=:}', indent=4)
                prints()
                prints('-------------------------------------------------', indent=4)
                prints(f'{ansi["yellow"]}Target Class:{ansi["reset"]}', indent=4)
                prints(f'{total_adv_target_conf=:}', indent=8)
                prints(f'{total_org_target_conf=:}', indent=8)
                prints(f'{succ_adv_target_conf=:}', indent=8)
                prints()
                prints('-------------------------------------------------', indent=4)
                prints(f'{ansi["yellow"]}Original Class:{ansi["reset"]}', indent=4)
                prints(f'{total_adv_org_conf=:}', indent=8)
                prints(f'{total_org_org_conf=:}', indent=8)
                prints(f'{succ_adv_org_conf=:}', indent=8)
        if verbose:
            prints(f'{ansi["green"]}{succ_iter_list.count} / {total_iter_list.count}{ansi["reset"]}')
            prints(succ_idx_list)
        if verbose >= 2:
            prints(f'{total_iter_list=:}', indent=4)
            prints(f'{succ_iter_list=:}', indent=4)
            prints()
            prints('-------------------------------------------------', indent=4)
            prints(f'{ansi["yellow"]}Target Class:{ansi["reset"]}', indent=4)
            prints(f'{total_adv_target_conf=:}', indent=8)
            prints(f'{total_org_target_conf=:}', indent=8)
            prints(f'{succ_adv_target_conf=:}', indent=8)
            prints()
            prints('-------------------------------------------------', indent=4)
            prints(f'{ansi["yellow"]}Original Class:{ansi["reset"]}', indent=4)
            prints(f'{total_adv_org_conf=:}', indent=8)
            prints(f'{total_org_org_conf=:}', indent=8)
            prints(f'{succ_adv_org_conf=:}', indent=8)
        return float(succ_iter_list.count) / total_iter_list.count, total_iter_list.global_avg

    def optimize(self, _input: torch.Tensor, *args,
                 target: None | int | torch.Tensor = None, target_idx: int = None,
                 loss_fn: Callable[..., torch.Tensor] = None,
                 require_class: bool = None,
                 loss_kwargs: dict[str, torch.Tensor] = {},
                 **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
        if len(_input) == 0:
            return _input, None
        target_idx = self.target_idx if target_idx is None else target_idx
        match target:
            case None:
                target = self.generate_target(_input, idx=self.target_idx) if self.target_class is None \
                    else self.target_class * torch.ones(len(_input), dtype=torch.long, device=_input.device)
            case int():
                target = target * torch.ones(len(_input), dtype=torch.long, device=_input.device)
        if loss_fn is None and self.loss_fn is None:
            untarget_condition = self.target_class is None and self.target_idx == 0

            def _loss_fn(_input: torch.Tensor, target: torch.Tensor, reduction: str = 'mean', **kwargs):
                loss = self.model.loss(_input, target, reduction=reduction)
                return -loss if untarget_condition else loss
            loss_fn = _loss_fn
        loss_kwargs.update(target=target)
        return super().optimize(_input, *args, target=target,
                                loss_fn=loss_fn, require_class=require_class,
                                loss_kwargs=loss_kwargs,
                                **kwargs)

    @torch.no_grad()
    def early_stop_check(self, current_idx: torch.Tensor,
                         adv_input: torch.Tensor, target: torch.Tensor, *args,
                         stop_threshold: float = None, require_class: bool = None,
                         **kwargs) -> torch.Tensor:
        stop_threshold = stop_threshold if stop_threshold is not None else self.stop_threshold
        require_class = require_class if require_class is not None else self.require_class
        if self.stop_threshold is None:
            return torch.zeros(len(current_idx), dtype=torch.bool)
        _confidence = self.model.get_target_prob(adv_input[current_idx], target[current_idx])
        untarget_condition = self.target_class is None and self.target_idx == 0
        result = _confidence > stop_threshold
        if untarget_condition:
            result = ~result
        if require_class:
            _class = self.model.get_class(adv_input[current_idx])
            class_result = _class == target[current_idx]
            if untarget_condition:
                class_result = ~class_result
            result = result.bitwise_and(class_result)
        return result

    @torch.no_grad()
    def output_info(self, org_input: torch.Tensor, noise: torch.Tensor, target: torch.Tensor,
                    loss_fn: Callable[[torch.Tensor], torch.Tensor] = None, **kwargs):
        super().output_info(org_input=org_input, noise=noise, loss_fn=loss_fn, **kwargs)
        # prints('Original class     : ', _label, indent=self.indent)
        # prints('Original confidence: ', _confidence, indent=self.indent)
        _confidence = self.model.get_target_prob(org_input + noise, target)
        prints('Target   class     : ', target.detach().cpu().tolist(), indent=self.indent)
        prints('Target   confidence: ', _confidence.detach().cpu().tolist(), indent=self.indent)



#!/usr/bin/env python3

from trojanvision.datasets.imagefolder import ImageFolder
from trojanvision.environ import env

import numpy as np
import os
import shutil
import pandas as pd
from tqdm import tqdm

from trojanvision import __file__ as root_file
root_dir = os.path.dirname(root_file)


class ISIC(ImageFolder):

    name: str = 'isic'

    def initialize_folder(self, **kwargs):
        super().initialize_folder(**kwargs)
        self.split_class('train')
        if self.valid_set:
            self.split_class('valid')

    def split_class(self, mode: str = 'train'):
        csv_path = os.path.normpath(os.path.join(root_dir, 'data', self.name, f'{mode}.csv'))
        print(f'Collect Label Information from CSV file: {csv_path}')
        obj = pd.read_csv(csv_path)
        labels = list(obj.columns.values)
        org_dict = {}
        for label in labels:
            org_dict[label] = np.array(obj[label]) if label == 'image' else np.array([
                bool(value) for value in obj[label]])
        new_dict = {}
        for label in labels[1:]:
            new_dict[label] = org_dict['image'][org_dict[label]]

        print('Splitting dataset to class folders ...')
        src_folder = os.path.normpath(os.path.join(self.folder_path, self.name, mode))
        if env['tqdm']:
            labels = tqdm(labels[1:], leave=False)
        for label in labels:
            seq = new_dict[label]
            dst_folder = os.path.join(src_folder, label)
            if not os.path.exists(dst_folder):
                os.makedirs(dst_folder)
            for img in seq:
                src = os.path.join(src_folder, img + '.jpg')
                dest = os.path.join(dst_folder, img + '.jpg')
                shutil.move(src, dest)


class ISIC2018(ISIC):
    r"""ISIC2018 dataset introduced by Noel Codella in 2018.
    It inherits :class:`trojanvision.datasets.ImageFolder`.

    See Also:
        * paper: `Skin Lesion Analysis Toward Melanoma Detection 2018\: A Challenge Hosted by the International Skin Imaging Collaboration (ISIC)`_
        * website: https://challenge.isic-archive.com/data/

    Attributes:
        name (str): ``'isic2018'``
        num_classes (int): ``7``
        data_shape (list[int]): ``[3, 224, 224]``

    .. _Skin Lesion Analysis Toward Melanoma Detection 2018\: A Challenge Hosted by the International Skin Imaging Collaboration (ISIC):
        https://arxiv.org/abs/1902.03368
    """  # noqa: E501

    name: str = 'isic2018'
    num_classes = 7
    url = {'train': 'https://isic-challenge-data.s3.amazonaws.com/2018/ISIC2018_Task3_Training_Input.zip',
           'valid': 'https://isic-challenge-data.s3.amazonaws.com/2018/ISIC2018_Task3_Validation_Input.zip'}
    md5 = {'train': '0c281f121070a8d63457caffcdec439a',
           'valid': 'c1fbdd4f5468b0d67c61a1b2def87077'}
    org_folder_name = {'train': 'ISIC2018_Task3_Training_Input',
                       'valid': 'ISIC2018_Task3_Validation_Input'}



#!/usr/bin/env python3

r"""
CUDA_VISIBLE_DEVICES=0 python examples/backdoor_attack.py --color --verbose 1 --pretrained --validate_interval 1 --lr 0.01 --mark_height 32 --mark_width 32 --attack invisible_poison --train_generator_epochs 10 --epochs 10 --mark_path tag_white.png
"""  # noqa: E501

from ...abstract import CleanLabelBackdoor
import trojanvision.models
from trojanzoo.environ import env
from trojanzoo.utils.logger import MetricLogger
from trojanzoo.utils.output import ansi, get_ansi_len, output_iter

import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision.models

from collections import OrderedDict
import os

import argparse


# https://github.com/junyanz/pytorch-CycleGAN-and-pix2pix/blob/master/models/networks.py
class ResNetBlock(nn.Module):
    def __init__(self, dim: int):
        super().__init__()
        self.conv_block = nn.Sequential(
            nn.ReflectionPad2d(1),
            # TODO: torchvision 0.13
            # torchvision.ops.Conv2dNormActivation(dim, dim, kernel_size=3, padding=0, bias=False),
            nn.Conv2d(dim, dim, kernel_size=3, bias=False),
            nn.BatchNorm2d(dim),
            nn.ReLU(True),
            nn.ReflectionPad2d(1),
            nn.Conv2d(dim, dim, kernel_size=3, bias=False),
            nn.BatchNorm2d(dim))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = x + self.conv_block(x)
        return out


class InvisiblePoison(CleanLabelBackdoor):
    r"""Invisible Poison Backdoor Attack proposed by Rui Ning
    from Old Dominion University in INFOCOM 2021.

    Based on :class:`trojanvision.attacks.CleanLabelBackdoor`,
    InvisiblePoison preprocesses the trigger by a generator (auto-encoder)
    to amplify its feature activation and make it invisible.

    See Also:
        * paper: `Invisible Poison\: A Blackbox Clean Label Backdoor Attack to Deep Neural Networks`_
        * code: https://github.com/rigley007/Invi_Poison

    Args:
        generator_mode (str): Choose from ``['default', 'resnet_comp', 'resnet']``
            Defaults to ``'default'``.
        noise_coeff (float): Minify rate of adversarial features.
            Defaults to ``0.35``.
        train_generator_epochs (int): Epochs of training generator (auto-encoder).
            Defaults to ``10``.

    .. _Invisible Poison\: A Blackbox Clean Label Backdoor Attack to Deep Neural Networks:
        https://ieeexplore.ieee.org/document/9488902
    """  # noqa: E501
    # noise_alpha (float): Weight of noise in :meth:`add_mark()`.
    #     Defaults to ``0.9``.
    name: str = 'invisible_poison'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--generator_mode',
                           help='(default: "default")')
        group.add_argument('--noise_coeff', type=float,
                           help='(default: 0.35)')
        group.add_argument('--noise_alpha', type=float,
                           help='weight of noise in add_mark '
                           '(default: 0.9)')
        group.add_argument('--train_generator_epochs', type=int,
                           help='epochs of training generator (auto-encoder) '
                           '(default: 10)')
        return group

    def __init__(self, generator_mode: str = 'default', noise_coeff: float = 0.35,
                 train_generator_epochs: int = 800, **kwargs):
        # noise_alpha: float = 0.9,
        super().__init__(**kwargs)
        assert self.mark.mark_height == self.dataset.data_shape[-2]
        assert self.mark.mark_width == self.dataset.data_shape[-1]
        self.mark.mark[-1] = 1.0

        self.param_list['invisible_poison'] = ['generator_mode', 'noise_coeff', 'train_generator_epochs']
        self.noise_coeff = noise_coeff
        self.train_generator_epochs = train_generator_epochs
        self.generator_mode = generator_mode
        # self.noise_alpha = noise_alpha

        self.generator = self.define_generator(generator_mode).to(device=env['device'])
        self.generator.requires_grad_(False)
        self.generator.eval()

    # def add_mark(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
    #     return (x + self.noise_alpha * self.mark.mark[:-1].to(device=x.device)).clamp(0, 1)

    def define_generator(self, mode: str = 'resnet') -> nn.Sequential:
        match mode:
            case 'resnet':
                # torchvision.models.feature_extraction.create_feature_extractor(
                #     torchvision.models.resnet18(pretrained=True),
                #     ['layer1'])
                resnet_model = torchvision.models.resnet18(pretrained=True)
                encoder = nn.Sequential(*list(resnet_model.children())[:5])
                bottleneck = nn.Sequential(
                    ResNetBlock(64),
                    ResNetBlock(64),
                    ResNetBlock(64))
                decoder = nn.Sequential(
                    nn.UpsamplingNearest2d(scale_factor=2),
                    nn.ConvTranspose2d(64, 3, kernel_size=7, stride=2, padding=3, output_padding=1, bias=False),
                    nn.Tanh())
            case 'resnet_comp':
                resnet_model = trojanvision.models.create('resnet18_comp',
                                                          dataset=self.dataset,
                                                          pretrained=True)
                encoder = nn.Sequential(*list(resnet_model._model.features.cpu().children())[:4])
                bottleneck = nn.Sequential(
                    ResNetBlock(64),
                    ResNetBlock(64),
                    ResNetBlock(64))
                decoder = nn.Sequential(
                    nn.ConvTranspose2d(64, 3, kernel_size=3, stride=1, padding=1, bias=False),
                    nn.Tanh())
            case 'default':
                encoder = nn.Sequential(
                    # MNIST:1*28*28
                    nn.Conv2d(3, 8, kernel_size=3, stride=1, padding=0, bias=True),
                    nn.InstanceNorm2d(8), nn.ReLU(),
                    # 8*26*26
                    nn.Conv2d(8, 16, kernel_size=3, stride=2, padding=0, bias=True),
                    nn.InstanceNorm2d(16), nn.ReLU(),
                    # 16*12*12
                    nn.Conv2d(16, 32, kernel_size=3, stride=2, padding=0, bias=True),
                    nn.InstanceNorm2d(32), nn.ReLU(),
                    # 32*5*5
                )
                bottleneck = nn.Sequential(
                    ResNetBlock(32),
                    ResNetBlock(32),
                    ResNetBlock(32),
                    ResNetBlock(32))
                decoder = nn.Sequential(
                    nn.ConvTranspose2d(32, 16, kernel_size=3, stride=2, padding=0, bias=False),
                    nn.InstanceNorm2d(16),
                    nn.ReLU(),
                    # state size. 16 x 11 x 11
                    nn.ConvTranspose2d(16, 8, kernel_size=3, stride=2, padding=0, bias=False),
                    nn.InstanceNorm2d(8),
                    nn.ReLU(),
                    # state size. 8 x 23 x 23
                    nn.ConvTranspose2d(8, 3, kernel_size=6, stride=1, padding=0, bias=False),
                    nn.Tanh()
                    # state size. image_nc x 28 x 28
                )
            case _:
                raise NotImplementedError(f'{self.generator_mode=}')
        return nn.Sequential(OrderedDict([
            ('encoder', encoder),
            ('bottleneck', bottleneck),
            ('decoder', decoder),
        ]))

    def attack(self, epochs: int, save: bool = False, **kwargs):
        # train generator
        resnet_model = trojanvision.models.create('resnet18_comp',
                                                  dataset=self.dataset,
                                                  pretrained=True)
        model_extractor = nn.Sequential(*list(resnet_model._model.features.children())[:4])
        match self.generator_mode:
            case 'resnet':
                resnet_model = torchvision.models.resnet18(pretrained=True).to(device=env['device'])
                model_extractor = nn.Sequential(*list(resnet_model.children())[:5])
            case _:
                resnet_model = trojanvision.models.create('resnet18_comp',
                                                          dataset=self.dataset,
                                                          pretrained=True)
                model_extractor = nn.Sequential(*list(resnet_model._model.features.children())[:4])
        model_extractor.requires_grad_(False)
        model_extractor.train()

        self.generator.train()
        if self.generator_mode == 'default':
            self.generator.requires_grad_()
            parameters = self.generator.parameters()
        else:
            self.generator.bottleneck.requires_grad_()
            self.generator.decoder.requires_grad_()
            parameters = self.generator[1:].parameters()
        optimizer = torch.optim.Adam(parameters, lr=1e-3)
        lr_scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
            optimizer, T_max=self.train_generator_epochs,
            eta_min=1e-5)

        logger = MetricLogger()
        logger.create_meters(loss=None, acc=None)
        for _epoch in range(self.train_generator_epochs):
            _epoch += 1
            logger.reset()
            header: str = '{blue_light}{0}: {1}{reset}'.format(
                'Epoch', output_iter(_epoch, self.train_generator_epochs), **ansi)
            header = header.ljust(max(len('Epoch'), 30) + get_ansi_len(header))
            loader = logger.log_every(self.dataset.loader['train'], header=header,
                                      tqdm_header='Batch')
            for data in loader:
                optimizer.zero_grad()
                _input, _label = self.model.get_data(data)
                adv_input = (self.generator(_input) + 1) / 2

                _feats = model_extractor(_input)
                adv_feats = model_extractor(adv_input)

                loss = F.l1_loss(_feats, self.noise_coeff * adv_feats)
                loss.backward()
                optimizer.step()
                batch_size = len(_label)
                with torch.no_grad():
                    org_class = self.model.get_class(_input)
                    adv_class = self.model.get_class(adv_input)
                    acc = (org_class == adv_class).float().sum().item() * 100.0 / batch_size
                logger.update(n=batch_size, loss=loss.item(), acc=acc)
            lr_scheduler.step()
        self.save_generator()
        optimizer.zero_grad()
        self.generator.eval()
        self.generator.requires_grad_(False)

        self.mark.mark[:-1] = (self.generator(self.mark.mark[:-1].unsqueeze(0))[0] + 1) / 2
        self.poison_set = self.get_poison_dataset(load_mark=False)
        return super().attack(epochs, save=save, **kwargs)

    def get_filename(self, **kwargs) -> str:
        return super().get_filename(**kwargs) + f'_{self.generator_mode}'

    def save_generator(self, filename: str = None, **kwargs):
        filename = filename or self.get_filename(**kwargs)
        file_path = os.path.join(self.folder_path, filename)
        torch.save(self.generator.state_dict(), file_path + '_generator.pth')
        print('generator results saved at: ', file_path)

    def load(self, filename: str = None, **kwargs):
        filename = filename or self.get_filename(**kwargs)
        file_path = os.path.join(self.folder_path, filename)
        self.generator.load_state_dict(torch.load(file_path + '_generator.pth'))
        return super().load(filename, **kwargs)



#!/usr/bin/env python3

from .abstract import BackdoorDefense, InputFiltering, ModelInspection, TrainingFiltering
from trojanzoo.defenses import Defense

from .adv import *
from .backdoor import *

from . import adv, backdoor

from trojanvision.configs import config
import trojanzoo.defenses

import argparse
from trojanzoo.configs import Config
from trojanvision.datasets import ImageSet

module_list = [adv, backdoor]
__all__ = ['add_argument', 'create',
           'Defense', 'BackdoorDefense',
           'InputFiltering', 'ModelInspection', 'TrainingFiltering']
class_dict: dict[str, type[Defense]] = {}
for module in module_list:
    __all__.extend(module.__all__)
    class_dict.update(module.class_dict)


def add_argument(parser: argparse.ArgumentParser,
                 defense_name: str = None, defense: str | Defense = None,
                 class_dict: dict[str, type[Defense]] = class_dict):
    return trojanzoo.defenses.add_argument(parser=parser, defense_name=defense_name, defense=defense,
                                           class_dict=class_dict)


def create(defense_name: str = None, defense: str | Defense = None,
           dataset_name: str = None, dataset: str | ImageSet = None,
           config: Config = config, class_dict: dict[str, type[Defense]] = class_dict, **kwargs):
    return trojanzoo.defenses.create(defense_name=defense_name, defense=defense,
                                     dataset_name=dataset_name, dataset=dataset,
                                     config=config, class_dict=class_dict, **kwargs)



#!/usr/bin/env python3

from .policy import Policy
from trojanvision.models import ImageModel

import torch
import torch.nn as nn
import torch.optim as optim


class WGAN(nn.Module):
    def __init__(self, model: ImageModel,
                 policy: Policy,
                 eps: float = 0.1,
                 gp_factor: float = 10.0):
        super().__init__()
        num_feats: int = model._model.classifier[0].in_features

        self.policy = policy
        self.model = model._model
        self.discriminator = nn.Sequential(
            nn.Linear(num_feats, num_feats),
            nn.ReLU(),
            nn.Linear(num_feats, 1))
        self.criterion = model.criterion

        self.eps = eps
        self.gp_factor = gp_factor

        self.model_optimizer = optim.Adam(model.parameters(),
                                          lr=1e-3, betas=[0.0, 0.999])
        self.policy_optimizer = optim.Adam(policy.parameters(),
                                           lr=1e-3, betas=[0.0, 0.999])

    def update(self, _input: torch.Tensor,
               _label: torch.Tensor) -> dict[str, float]:
        a_input, n_input = _input.chunk(2)
        a_label, n_label = _label.chunk(2)

        # Discriminator Update
        with torch.no_grad():
            aug_input = self.policy(a_input)
        n_feats = self.model.get_final_fm(n_input)
        n_output = self.model.classifier(n_feats)
        cls_n_loss: torch.Tensor = self.criterion(n_output, n_label)
        d_n_output: torch.Tensor = self.discriminator(n_feats)
        d_n_loss = d_n_output.mean()

        a_feats = self.model.get_final_fm(aug_input)
        d_a_output: torch.Tensor = self.discriminator(a_feats)
        d_a_loss = d_a_output.mean()

        gp_loss = self.gradient_penalty(n_input, aug_input)

        d_loss = self.eps * cls_n_loss - d_n_loss + d_a_loss \
            + self.gp_factor * gp_loss
        self.model_optimizer.zero_grad()
        d_loss.backward()
        self.model_optimizer.step()

        # Generator Update
        aug2_input = self.policy(a_input)
        a2_feats = self.model.get_final_fm(aug2_input)
        a2_output = self.model.classifier(a2_feats)
        cls_a2_loss: torch.Tensor = self.criterion(a2_output, a_label)

        d_a2_output: torch.Tensor = self.discriminator(a2_feats)
        d_a2_loss = d_a2_output.mean()

        g_loss = self.eps * cls_a2_loss - d_a2_loss
        self.policy_optimizer.zero_grad()
        g_loss.backward()
        self.policy_optimizer.step()
        for param in self.policy.parameters():
            param.data.nan_to_num_(0.5)

        loss_dict: dict[str, float] = {
            'discriminator': d_loss.item(),
            'generator': g_loss.item(),
            'classification': (cls_n_loss + cls_a2_loss).item(),
            'gradient_penalty': gp_loss.item()
        }
        return loss_dict

    def gradient_penalty(self, real: torch.Tensor,
                         fake: torch.Tensor) -> torch.Tensor:
        alpha = torch.rand(real.size(0), device=real.device)
        alpha = alpha.view([-1] + [1] * (real.dim() - 1))
        interpolated: torch.Tensor = alpha * real + (1 - alpha) * fake
        interpolated = interpolated.detach()
        interpolated.requires_grad_()

        _feats = self.model.get_final_fm(interpolated)
        d_output: torch.Tensor = self.discriminator(_feats)
        grad = torch.autograd.grad(d_output.sum(), interpolated,
                                   create_graph=True)[0]
        result: torch.Tensor = grad.norm(2, dim=1) - 1
        return result.square().mean()



#!/usr/bin/env python3

from trojanvision import environ as environ
from trojanvision import datasets as datasets
from trojanvision import models as models
from trojanvision import trainer as trainer

from trojanvision import attacks as attacks
from trojanvision import defenses as defenses
from trojanvision import marks as marks

from trojanzoo import __version__
from trojanzoo import summary

__all__ = ['summary']



#!/usr/bin/env python3

r"""--nats_path /data/rbp5354/nats/NATS-tss-v1_0-3ffb9-full"""  # noqa: E501

from trojanvision.datasets.imageset import ImageSet
from trojanvision.models.imagemodel import _ImageModel, ImageModel

import torch
import torch.nn as nn
from collections import OrderedDict

import argparse
from typing import Any
from collections.abc import Callable


class _NATSbench(_ImageModel):

    def __init__(self, network: nn.Module = None, **kwargs):
        super().__init__(**kwargs)
        self.load_model(network)

    def load_model(self, network: nn.Module):
        self.features = nn.Sequential(OrderedDict([
            ('stem', getattr(network, 'stem')),
            ('cells', nn.Sequential(*getattr(network, 'cells'))),
            ('lastact', getattr(network, 'lastact')),
        ]))
        self.classifier = nn.Sequential(OrderedDict([
            ('fc', getattr(network, 'classifier'))
        ]))


class NATSbench(ImageModel):
    r"""NATS-Bench proposed by Xuanyi Dong from University of Technology Sydney.

    :Available model names:

        .. code-block:: python3

            ['nats_bench']

    Note:
        There are prerequisites to use the benchmark:

        * ``pip install nats_bench``.
        * ``git clone https://github.com/D-X-Y/AutoDL-Projects.git`` and ``pip install .``
        * Extract ``NATS-tss-v1_0-3ffb9-full`` to :attr:`nats_path`.

    See Also:

        * paper: `NATS-Bench\: Benchmarking NAS Algorithms for Architecture Topology and Size`_
        * code:

          - AutoDL: https://github.com/D-X-Y/AutoDL-Projects
          - NATS-Bench: https://github.com/D-X-Y/NATS-Bench

    Args:
        model_index (int): :attr:`model_index` passed to
            ``api.get_net_config()``.
            Ranging from ``0 -- 15624``.
            Defaults to ``0``.
        model_seed (int): :attr:`model_seed` passed to
            ``api.get_net_param()``.
            Choose from ``[777, 888, 999]``.
            Defaults to ``999``.
        hp (int): Training epochs.
            :attr:`hp` passed to ``api.get_net_param()``.
            Choose from ``[12, 200]``.
            Defaults to ``200``.
        nats_path (str): NATS benchmark file path.
            It should be set as format like
            ``'**/NATS-tss-v1_0-3ffb9-full'``
        search_space (str): Search space of topology or size.
            Choose from ``['tss', 'sss']``.
        dataset_name (str): Dataset name.
            Choose from ``['cifar10', 'cifar10-valid', 'cifar100', 'imagenet16-120']``.

    .. _NATS-Bench\: Benchmarking NAS Algorithms for Architecture Topology and Size:
        https://arxiv.org/abs/2009.00437
    """
    available_models = ['nats_bench']

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--model_index', type=int)
        group.add_argument('--model_seed', type=int)
        group.add_argument('--hp', type=int)
        group.add_argument('--nats_path')
        group.add_argument('--search_space')
        return group

    def __init__(self, name: str = 'nats_bench', model: type[_NATSbench] = _NATSbench,
                 model_index: int = 0, model_seed: int = 999, hp: int = 200,
                 dataset: ImageSet | None = None, dataset_name: str | None = None,
                 nats_path: str | None = None,
                 search_space: str = 'tss', **kwargs):
        try:
            # pip install nats_bench
            from nats_bench import create   # type: ignore
            from xautodl.models import get_cell_based_tiny_net   # type: ignore
        except ImportError:
            raise ImportError('You need to install nats_bench and auto-dl library')

        if isinstance(dataset, ImageSet):
            kwargs['dataset'] = dataset
            if dataset_name is None:
                dataset_name = dataset.name
                if dataset_name == 'imagenet16':
                    dataset_name = f'imagenet16-{dataset.num_classes:d}'
        assert dataset_name is not None
        dataset_name = dataset_name.replace('imagenet16', 'ImageNet16')
        self.dataset_name = dataset_name

        self.model_index = model_index
        self.model_seed = model_seed
        self.hp = hp
        self.search_space = search_space
        self.nats_path = nats_path

        self.api = create(nats_path, search_space, fast_mode=True, verbose=False)
        config: dict[str, Any] = self.api.get_net_config(model_index, dataset_name)
        self.get_cell_based_tiny_net: Callable[..., nn.Module] = get_cell_based_tiny_net
        network = self.get_cell_based_tiny_net(config)
        super().__init__(name=name, model=model, network=network, **kwargs)
        self.param_list['nats_bench'] = ['model_index', 'model_seed', 'hp', 'search_space', 'nats_path']
        self._model: _NATSbench

    def get_official_weights(self, model_index: int | None = None,
                             model_seed: int | None = None,
                             hp: int | None = None,
                             **kwargs) -> OrderedDict[str, torch.Tensor]:
        model_index = model_index if model_index is not None else self.model_index
        model_seed = model_seed if model_seed is not None else self.model_seed
        hp = hp if hp is not None else self.hp
        _dict: OrderedDict[str, torch.Tensor] = self.api.get_net_param(
            model_index, self.dataset_name, model_seed, hp=str(hp))
        if _dict is None:
            raise FileNotFoundError(f'Loaded weight is None. Please check {self.nats_path=}.\n'
                                    'It should be set as format like "**/NATS-tss-v1_0-3ffb9-full"``')
        new_dict: OrderedDict[str, torch.Tensor] = OrderedDict()
        for k, v in _dict.items():
            if k.startswith('stem') or k.startswith('cells') or k.startswith('lastact'):
                new_dict['features.' + k] = v
            elif k.startswith('classifier'):
                new_dict['classifier.fc' + k[10:]] = v
        return new_dict



#!/usr/bin/env python3

from ...abstract import BackdoorAttack
from trojanzoo.utils.model import init_weights
from trojanvision.attacks.adv import PGD    # TODO: Need to check whether this will cause ImportError
from trojanvision.optim import PGDoptimizer
from trojanvision.environ import env
from trojanzoo.utils.data import TensorListDataset, dataset_to_tensor, sample_batch

import torch
import torch.nn as nn
import torch.optim as optim

import argparse
import os

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    import torch.utils.data


class LabelConsistent(BackdoorAttack):
    r"""
    Contributor: Xiangshan Gao, Ren Pang

    Clean Label Backdoor Attack is described in detail in the paper `Clean Label Backdoor Attack`_ by Alexander Turner.

    The main idea is to perturb the poisoned samples
    in order to render learning the salient characteristic of the input more difficult,
    causing the model rely more heavily on the backdoor pattern in order to successfully introduce backdoor.

    Utilize the adversarial examples and GAB generated data,
    the resulted poisoned inputs appear to be consistent with their label
    and thus seem benign even upon human inspection.

    The authors haven't posted `original source code`_.

    Args:
        poison_generation_method (str): the chosen method to generate poisoned sample. Default: 'pgd'.
        tau (float): the interpolation constant used to balance source imgs and target imgs. Default: 0.4.
        epsilon (float): the perturbation bound in input space. Default: 0.1.
        noise_dim (int): the dimension of the input in the generator. Default: 100.
        generator_iters (int): the epochs for training the generator. Default: 1000.
        critic_iter (int): the critic iterations per generator training iteration. Default: 5.


    .. _Clean Label:
        https://people.csail.mit.edu/madry/lab/cleanlabel.pdf

    .. _related code:
        https://github.com/igul222/improved_wgan_training
        https://github.com/MadryLab/cifar10_challenge
        https://github.com/caogang/wgan-gp

    """
    name: str = 'label_consistent'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--poison_generation_method', choices=['pgd', 'gan'],
                           help='the chosen method to generate poisoned sample '
                           '(default: "pgd")')
        group.add_argument('--pgd_alpha', type=float)
        group.add_argument('--pgd_eps', type=float)
        group.add_argument('--pgd_iter', type=int)
        group.add_argument('--tau', type=float,
                           help='the interpolation constant used to balance source imgs and target imgs, '
                           'defaults to 0.2')
        group.add_argument('--noise_dim', type=int,
                           help='the dimension of the input in the generator, '
                           'defaults to config[clean_label][noise_dim]=100')
        group.add_argument('--train_gan', action='store_true',
                           help='whether train the GAN if it already exists, defaults to False')
        group.add_argument('--generator_iters', type=int,
                           help='epochs for training the generator, defaults to 1000')
        group.add_argument('--critic_iter', type=int,
                           help='critic iterations per generator training iteration '
                           '(default: 5)')
        return group

    def __init__(self, poison_generation_method: str = 'pgd',
                 pgd_alpha: float = 2 / 255, pgd_eps: float = 8 / 255, pgd_iter=7,
                 tau: float = 0.2, noise_dim: int = 100,
                 train_gan: bool = False, generator_iters: int = 1000, critic_iter: int = 5, **kwargs):
        super().__init__(**kwargs)
        self.param_list['clean_label'] = ['poison_generation_method', 'poison_num']
        self.poison_generation_method: str = poison_generation_method
        match poison_generation_method:
            case 'pgd':
                self.poison_num: int = int(len(self.dataset.get_dataset(
                    'train', class_list=[self.target_class])) * self.poison_percent)
            case 'gan':
                self.poison_num: int = int(len(self.dataset.get_dataset('train')) * self.poison_percent)

        match poison_generation_method:
            case 'pgd':
                self.param_list['pgd'] = ['pgd_alpha', 'pgd_eps', 'pgd_iter']
                if pgd_alpha is None:
                    pgd_alpha = 1.5 * pgd_eps / pgd_iter
                self.pgd_alpha: float = pgd_alpha
                self.pgd_eps: float = pgd_eps
                self.pgd_iter: int = pgd_iter
                self.pgd: PGD = PGD(pgd_alpha=pgd_alpha, pgd_eps=pgd_eps, iteration=pgd_iter,
                                    target_idx=0, output=self.output, dataset=self.dataset, model=self.model)
            case 'gan':
                self.param_list['gan'] = ['tau', 'noise_dim', 'train_gan', 'critic_iter', 'generator_iters']
                self.tau: float = tau
                self.noise_dim: int = noise_dim
                self.train_gan: bool = train_gan
                self.generator_iters = generator_iters
                self.critic_iter = critic_iter
                self.wgan = WGAN(noise_dim=self.noise_dim, dim=64, data_shape=self.dataset.data_shape,
                                 generator_iters=self.generator_iters, critic_iter=self.critic_iter)

    def attack(self, optimizer: torch.optim.Optimizer, lr_scheduler: torch.optim.lr_scheduler._LRScheduler, **kwargs):

        target_class_set = self.dataset.get_dataset('train', class_list=[self.target_class])
        target_imgs, _ = sample_batch(target_class_set, batch_size=self.poison_num)
        target_imgs = target_imgs.to(env['device'])

        full_set = self.dataset.get_dataset('train')
        poison_set: TensorListDataset = None    # TODO
        match self.poison_generation_method:
            case 'pgd':
                trigger_label = self.target_class * torch.ones(
                    len(target_imgs), dtype=torch.long, device=target_imgs.device)
                result = []
                for data in zip(target_imgs.chunk(self.dataset.batch_size),
                                trigger_label.chunk(self.dataset.batch_size)):
                    poison_img, _ = self.model.remove_misclassify(data)
                    poison_img, _ = self.pgd.optimize(poison_img)
                    poison_img = self.add_mark(poison_img).cpu()
                    result.append(poison_img)
                poison_imgs = torch.cat(result)

                poison_set = TensorListDataset(poison_imgs, [self.target_class] * len(poison_imgs))
                # poison_set = torch.utils.data.ConcatDataset([poison_set, target_original_dataset])

            case 'gan':
                other_classes = list(range(self.dataset.num_classes))
                other_classes.pop(self.target_class)
                x_list = []
                y_list = []
                for source_class in other_classes:
                    print('Process data of Source Class: ', source_class)
                    source_class_dataset = self.dataset.get_dataset(mode='train', class_list=[source_class])
                    sample_source_class_dataset, _ = self.dataset.split_dataset(
                        source_class_dataset, self.poison_num)
                    source_imgs = dataset_to_tensor(sample_source_class_dataset)[0].to(device=env['device'])

                    g_path = f'{self.folder_path}gan_dim{self.noise_dim}_class{source_class}_g.pth'
                    d_path = f'{self.folder_path}gan_dim{self.noise_dim}_class{source_class}_d.pth'
                    if os.path.exists(g_path) and os.path.exists(d_path) and not self.train_gan:
                        self.wgan.G.load_state_dict(torch.load(g_path, map_location=env['device']))
                        self.wgan.D.load_state_dict(torch.load(d_path, map_location=env['device']))
                        print(f'    load model from: \n        {g_path}\n        {d_path}', )
                    else:
                        self.train_gan = True
                        self.wgan.reset_parameters()
                        gan_dataset = torch.utils.data.ConcatDataset([source_class_dataset, target_class_set])
                        gan_dataloader = self.dataset.get_dataloader(
                            mode='train', dataset=gan_dataset, batch_size=self.dataset.batch_size, num_workers=0)
                        self.wgan.train(gan_dataloader)
                        torch.save(self.wgan.G.state_dict(), g_path)
                        torch.save(self.wgan.D.state_dict(), d_path)
                        print(f'GAN Model Saved at : \n{g_path}\n{d_path}')
                        continue

                    for source_chunk, target_chunk in zip(source_imgs.chunk(self.dataset.batch_size),
                                                          target_imgs.chunk(self.dataset.batch_size)):
                        source_encode = self.wgan.get_encode_value(source_chunk).detach()
                        target_encode = self.wgan.get_encode_value(target_chunk).detach()
                        # noise = torch.randn_like(source_encode)
                        # source_img = self.wgan.G(source_encode)
                        # target_img = self.wgan.G(target_encode)
                        # if not os.path.exists('./imgs'):
                        #     os.makedirs('./imgs')
                        # for i in range(len(source_img)):
                        #     F.to_pil_image(source_img[i]).save(f'./imgs/source_{i}.png')
                        # for i in range(len(target_img)):
                        #     F.to_pil_image(target_img[i]).save(f'./imgs/target_{i}.png')
                        # exit()
                        interpolation_encode = source_encode * self.tau + target_encode * (1 - self.tau)
                        poison_imgs = self.wgan.G(interpolation_encode).detach()
                        poison_imgs = self.add_mark(poison_imgs)

                        poison_imgs = poison_imgs.cpu()
                        x_list.append(poison_imgs)
                    y_list.extend([self.target_class] * len(source_imgs))
                assert not self.train_gan
                x_list = torch.cat(x_list)
                poison_set = TensorListDataset(x_list, y_list)
                # poison_set = torch.utils.data.ConcatDataset([poison_set, target_original_dataset])
        final_set = torch.utils.data.ConcatDataset([poison_set, full_set])
        # final_set = poison_set
        final_loader = self.dataset.get_dataloader(mode='train', dataset=final_set, num_workers=0)
        self.model._train(optimizer=optimizer, lr_scheduler=lr_scheduler, save_fn=self.save,
                          loader_train=final_loader, validate_fn=self.validate_fn, **kwargs)


class Generator(nn.Module):
    def __init__(self, noise_dim: int = 100, dim: int = 64, data_shape: list[int] = [3, 32, 32]):
        super().__init__()
        self.noise_dim: int = noise_dim
        self.dim: int = dim
        self.data_shape: list[int] = data_shape
        init_dim = dim * data_shape[1] * data_shape[2] // 16
        self.preprocess = nn.Linear(noise_dim, init_dim)
        self.preprocess_1 = nn.Sequential(
            nn.BatchNorm2d(init_dim),
            nn.ReLU(True),)
        self.block1 = nn.Sequential(
            nn.ConvTranspose2d(4 * dim, 2 * dim, 2, stride=2),
            nn.BatchNorm2d(2 * dim),
            nn.ReLU(True),)
        self.block2 = nn.Sequential(
            nn.ConvTranspose2d(2 * dim, dim, 2, stride=2),
            nn.BatchNorm2d(dim),
            nn.ReLU(True),)
        self.deconv_out = nn.ConvTranspose2d(dim, data_shape[0], 2, stride=2)
        self.tanh = nn.Tanh()

    def forward(self, x: torch.Tensor):
        # (N, noise_dim)
        x = self.preprocess(x)
        # (N, noise_dim)
        x = x.unsqueeze(-1).unsqueeze(-1)
        x = self.preprocess_1(x)
        x = x.view(len(x), 4 * self.dim, self.data_shape[1] // 8, self.data_shape[2] // 8)
        x = self.block1(x)
        x = self.block2(x)
        x = self.deconv_out(x)
        x = self.tanh(x)
        return x


class Discriminator(nn.Module):
    def __init__(self, dim: int = 64, data_shape: list = [3, 32, 32]):
        super(Discriminator, self).__init__()
        self.dim = dim
        self.main = nn.Sequential(
            nn.Conv2d(data_shape[0], dim, 3, 2, padding=1),
            nn.LeakyReLU(),
            nn.Conv2d(dim, 2 * dim, 3, 2, padding=1),
            nn.BatchNorm2d(2 * dim),
            nn.LeakyReLU(),
            nn.Conv2d(2 * dim, 4 * dim, 3, 2, padding=1),
            nn.BatchNorm2d(4 * dim),
            nn.LeakyReLU(),
        )
        init_dim = dim * data_shape[1] * data_shape[2] // 16
        self.linear = nn.Linear(init_dim, 1)

    def forward(self, x: torch.Tensor):
        x = self.main(x)
        x = x.flatten(start_dim=1)
        x = self.linear(x)
        return x


class WGAN(object):
    def __init__(self, noise_dim: int, dim: int, data_shape: list[int] = [3, 32, 32],
                 generator_iters: int = 1000, critic_iter: int = 5):
        self.noise_dim = noise_dim
        self.G: Generator = Generator(noise_dim, dim, data_shape)
        self.D: Discriminator = Discriminator(dim, data_shape)
        if env['num_gpus']:
            self.G.cuda()
            self.D.cuda()
        # the parameter in the original paper
        self.d_optimizer = optim.RMSprop(self.D.parameters(), lr=5e-5)
        self.g_optimizer = optim.RMSprop(self.G.parameters(), lr=5e-5)
        self.generator_iters = generator_iters  # larger: 1000
        self.critic_iter = critic_iter
        self.mse_loss = torch.nn.MSELoss()

        self.gan_pgd = PGDoptimizer(pgd_eps=1.0, iteration=500, output=0)

    def reset_parameters(self):
        init_weights(self.G)
        init_weights(self.D)

    def train(self, train_dataloader):
        self.g_optimizer.zero_grad()
        self.d_optimizer.zero_grad()
        for g_iter in range(self.generator_iters):
            # Requires grad, Generator requires_grad = False
            for p in self.D.parameters():
                p.requires_grad = True
                p.data.clamp_(-0.01, 0.01)
            for p in self.G.parameters():
                p.requires_grad = False

            for d_iter in range(self.critic_iter):
                d_loss_fake = 0    # TODO
                d_loss_real = 0    # TODO
                for i, (data, label) in enumerate(train_dataloader):
                    data = torch.tensor(data)
                    train_data = data.to(env['device'])
                    d_loss_real = self.D(train_data).mean()

                    z = torch.randn(train_data.shape[0], self.noise_dim, device=train_data.device)
                    fake_images = self.G(z)
                    d_loss_fake = self.D(fake_images).mean()

                    d_loss = d_loss_fake - d_loss_real
                    d_loss.backward()
                    self.d_optimizer.step()
                    self.d_optimizer.zero_grad()
                print(f'    Discriminator: loss_fake: {d_loss_fake:.5f}, loss_real: {d_loss_real:.5f}')
            for p in self.D.parameters():
                p.requires_grad = False
            for p in self.G.parameters():
                p.requires_grad = True
            g_loss = 0    # TODO
            for i, (data, label) in enumerate(train_dataloader):
                data = torch.tensor(data)
                train_data = data.to(env['device'])
                z = torch.randn(train_data.shape[0], self.noise_dim, device=train_data.device)
                fake_images = self.G(z)
                g_loss = - self.D(fake_images).mean()
                g_loss.backward()
                self.g_optimizer.step()
                self.g_optimizer.zero_grad()
            print(f'Generator iteration: {g_iter:5d} / {self.generator_iters:5d}, g_loss: {g_loss:.5f}')

    def get_encode_value(self, imgs: torch.Tensor):
        """According to the image and Generator, utilize pgd optimization to get the d dimension encoding value.

        Args:
            imgs (torch.Tensor): the chosen image to get its encoding value, also considered as the output of Generator.
            noise_dim (int): the dimension of the input in the generator.

        Returns:
            torch.Tensor: the synthesized poisoned image.
        """

        def loss_func(X: torch.Tensor):
            loss = self.mse_loss(self.G(X), imgs)
            return loss
        x_1 = torch.randn(len(imgs), self.noise_dim, device=imgs.device)
        x_1, _ = self.gan_pgd.optimize(_input=x_1, loss_fn=loss_func)
        return x_1



#!/usr/bin/env python3

r"""
CUDA_VISIBLE_DEVICES=0 python examples/backdoor_attack.py --color --verbose 1 --pretrained --validate_interval 1 --epochs 20 --lr 0.01 --attack clean_label --tqdm
"""  # noqa: E501

from ...abstract import CleanLabelBackdoor

import torch.utils.data


class CleanLabel(CleanLabelBackdoor):
    def __init__(self, *args, train_mode: str = 'dataset', **kwargs):
        super().__init__(*args, train_mode=train_mode, **kwargs)
        self.poison_set = self.get_poison_dataset()

    def get_poison_dataset(self, poison_num: int = None,
                           seed: int = None
                           ) -> torch.utils.data.Dataset:
        return super().get_poison_dataset(poison_num, load_mark=False, seed=seed)



#!/usr/bin/env python3

from trojanzoo.environ import env
from trojanzoo.utils.output import ansi, prints
from trojanzoo.utils.tensor import repeat_to_batch

import torch
import torch.nn as nn
import torchvision.transforms as transforms

from collections.abc import Callable, Iterable
from typing import Iterator

__all__ = ['get_all_layer', 'get_layer', 'get_layer_name',
           'summary', 'activate_params', 'accuracy', 'generate_target']

filter_tuple: tuple[nn.Module] = (transforms.Normalize, nn.Dropout,
                                  nn.BatchNorm2d, nn.GroupNorm,
                                  nn.ReLU, nn.Sigmoid)


def init_weights(m: nn.Module, filter_list: list[type] = []) -> None:
    r"""Traverse module :attr:`m` to intialize weights of all submodules
    except for those in :attr:`filter_list`.

    Module parameters are reset by calling ``module.reset_parameters()``.

    Note:
        An alternative implementation is to call ``m.apply(init_weights)``
        so that this method could be non-recursive and avoid traverse.

    Args:
        m (torch.nn.Module): Module to initialize.
        filter_list (tuple[type]): List of submodule types as exceptions.
            Defaults to ``[]`` (empty).

    :Example:
        .. code-block:: python
            :emphasize-lines: 5-6

            from trojanzoo.utils.model import init_weights
            import torch.nn as nn

            net = nn.Sequential(nn.Linear(2, 2), nn.Linear(2, 2))
            init_weights(filter_list=[nn.Linear])   # no change
            init_weights(net)                       # init nn.Linear layers
    """
    if isinstance(m, tuple(filter_list)):
        return
    if hasattr(m, 'reset_parameters'):
        return m.reset_parameters()
    else:
        for layer in m.children():
            init_weights(layer, filter_list=filter_list)


def get_layer_name(module: nn.Module, depth: int = -1, prefix: str = '',
                   use_filter: bool = True, non_leaf: bool = False,
                   seq_only: bool = False, init: bool = True) -> list[str]:
    r"""Get layer names of a :any:`torch.nn.Module`.

    Args:
        module (torch.nn.Module): the module to process.
        depth (int): The traverse depth.
            Defaults to ``-1`` (:math:`\infty`).
        prefix (str): The prefix string to all elements.
            Defaults to empty string ``''``.
        use_filter (bool): Whether to filter out certain layer types.

            * :any:`torchvision.transforms.Normalize`
            * :any:`torch.nn.Dropout`
            * :any:`torch.nn.BatchNorm2d`
            * :any:`torch.nn.ReLU`
            * :any:`torch.nn.Sigmoid`
        non_leaf (bool): Whether to include non-leaf nodes.
            Defaults to ``False``.
        seq_only (bool): Whether to only traverse children
            of :any:`torch.nn.Sequential`.
            If ``False``, will traverse children of all :any:`torch.nn.Module`.
            Defaults to ``False``.

    Returns:
        list[str]: The list of all layer names.

    :Example:
        >>> import torchvision
        >>> from trojanzoo.utils.model import get_layer_name
        >>>
        >>> model = torchvision.models.resnet18()
        >>> get_layer_name(model, depth=0)
        []
        >>> get_layer_name(model, depth=1)
        ['conv1', 'maxpool', 'layer1', 'layer2',
        'layer3', 'layer4', 'avgpool', 'fc']
        >>> get_layer_name(model, depth=2, prefix='model')
        ['model.conv1', 'model.maxpool', 'model.layer1.0', 'model.layer1.1',
        'model.layer2.0', 'model.layer2.1', 'model.layer3.0', 'model.layer3.1',
        'model.layer4.0', 'model.layer4.1', 'model.avgpool', 'model.fc']
        >>> get_layer_name(model, seq_only=True)
        ['conv1', 'maxpool', 'layer1.0', 'layer1.1', 'layer2.0', 'layer2.1',
        'layer3.0', 'layer3.1', 'layer4.0', 'layer4.1', 'avgpool', 'fc']
        >>> get_layer_name(model, seq_only=True, non_leaf=True)
        ['conv1', 'maxpool',
        'layer1.0', 'layer1.1', 'layer1',
        'layer2.0', 'layer2.1', 'layer2',
        'layer3.0', 'layer3.1', 'layer3',
        'layer4.0', 'layer4.1', 'layer4',
        'avgpool', 'fc']
        >>> get_layer_name(model)
        ['conv1', 'maxpool',
        'layer1.0.conv1', 'layer1.0.conv2', 'layer1.1.conv1', 'layer1.1.conv2',
        'layer2.0.conv1', 'layer2.0.conv2', 'layer2.0.downsample.0', 'layer2.1.conv1', 'layer2.1.conv2',
        'layer3.0.conv1', 'layer3.0.conv2', 'layer3.0.downsample.0', 'layer3.1.conv1', 'layer3.1.conv2',
        'layer4.0.conv1', 'layer4.0.conv2', 'layer4.0.downsample.0', 'layer4.1.conv1', 'layer4.1.conv2',
        'avgpool', 'fc']
    """  # noqa: E501
    layer_name_list: list[str] = []
    is_leaf = True
    if (init or (not seq_only or isinstance(module, nn.Sequential)))\
            and depth != 0:
        for name, child in module.named_children():
            full_name = prefix + ('.' if prefix else '') + name  # prefix=full_name
            layer_name_list.extend(get_layer_name(child, depth - 1, full_name,
                                                  use_filter, non_leaf, seq_only,
                                                  init=False))
            is_leaf = False
    if prefix and (not use_filter or filter_layer(module)) \
            and (non_leaf or depth == 0 or is_leaf):
        layer_name_list.append(prefix)
    return layer_name_list


def get_all_layer(module: nn.Module, x: torch.Tensor,
                  layer_input: str = 'input', depth: int = -1,
                  prefix='', use_filter: bool = True, non_leaf: bool = False,
                  seq_only: bool = True, verbose: int = 0
                  ) -> dict[str, torch.Tensor]:
    r"""Get all intermediate layer outputs of
    :attr:`_input` from any intermediate layer
    in a :any:`torch.nn.Module`.

    Args:
        module (torch.nn.Module): the module to process.
        x (torch.Tensor): The batched input tensor
            from :attr:`layer_input`.
        layer_input (str): The intermediate layer name of :attr:`x`.
            Defaults to ``'input'``.
        depth (int): The traverse depth.
            Defaults to ``-1`` (:math:`\infty`).
        prefix (str): The prefix string to all elements.
            Defaults to empty string ``''``.
        use_filter (bool): Whether to filter out certain layer types.

            * :any:`torchvision.transforms.Normalize`
            * :any:`torch.nn.Dropout`
            * :any:`torch.nn.BatchNorm2d`
            * :any:`torch.nn.ReLU`
            * :any:`torch.nn.Sigmoid`
        non_leaf (bool): Whether to include non-leaf nodes.
            Defaults to ``False``.
        seq_only (bool): Whether to only traverse children
            of :any:`torch.nn.Sequential`.
            If ``False``, will traverse children of all :any:`torch.nn.Module`.
            Defaults to ``False``.
        verbose (int): The output level to show information
            including layer name, output shape and module information.
            Setting it larger than ``0`` will enable the output.
            Different integer values stands for different module information.
            Defaults to ``0``.

            * ``0``: No output
            * ``1``: Show layer class name.
            * ``2``: Show layer string (first line).
            * ``3``: Show layer string (full).

    Returns:
        dict[str, torch.Tensor]: The dict of all layer outputs.

    :Example:
        >>> import torch
        >>> import torchvision
        >>> from trojanzoo.utils.model import get_all_layer
        >>>
        >>> model = torchvision.models.densenet121()
        >>> x = torch.randn(5, 3, 224, 224)
        >>> y = get_all_layer(model.features, x, verbose=True)
        layer name                                        output shape        module information
        conv0                                             [5, 64, 112, 112]   Conv2d
        pool0                                             [5, 64, 56, 56]     MaxPool2d
        denseblock1                                       [5, 256, 56, 56]    _DenseBlock
        transition1.conv                                  [5, 128, 56, 56]    Conv2d
        transition1.pool                                  [5, 128, 28, 28]    AvgPool2d
        denseblock2                                       [5, 512, 28, 28]    _DenseBlock
        transition2.conv                                  [5, 256, 28, 28]    Conv2d
        transition2.pool                                  [5, 256, 14, 14]    AvgPool2d
        denseblock3                                       [5, 1024, 14, 14]   _DenseBlock
        transition3.conv                                  [5, 512, 14, 14]    Conv2d
        transition3.pool                                  [5, 512, 7, 7]      AvgPool2d
        denseblock4                                       [5, 1024, 7, 7]     _DenseBlock
        >>> y.keys()
        dict_keys(['conv0', 'pool0',
        'denseblock1', 'transition1.conv', 'transition1.pool',
        'denseblock2', 'transition2.conv', 'transition2.pool',
        'denseblock3', 'transition3.conv', 'transition3.pool',
        'denseblock4'])

    Note:
        This method regards :attr:`module` as a :any:`torch.nn.Sequential`.
        Many modules embed flatten operation in their :attr:`forward` method
        (e.g., ``view(n, -1)`` or ``flatten(1)``),
        making ``get_all_layer`` raise error.
        We suggest to use :any:`torch.nn.Flatten` instead to keep the module sequential.
    """  # noqa: E501
    layer_name_list = get_layer_name(
        module, depth=depth, prefix=prefix, use_filter=False)
    if layer_input == 'input':
        layer_input = 'record'
    elif layer_input not in layer_name_list:
        print('Model Layer Name List: ', layer_name_list)
        print('Input layer: ', layer_input)
        raise ValueError('Layer name not in model')
    if verbose:
        print(f'{ansi["green"]}{"layer name":<50s}'
              f'{"output shape":<20}{"module information"}{ansi["reset"]}')
    return _get_all_layer(module, x, layer_input, depth,
                          prefix, use_filter, non_leaf,
                          seq_only, verbose=verbose, init=True)[0]


def _get_all_layer(module: nn.Module, x: torch.Tensor,
                   layer_input: str = 'record', depth: int = 0,
                   prefix: str = '', use_filter: bool = True,
                   non_leaf: bool = False, seq_only: bool = True,
                   verbose: int = 0, init: bool = False
                   ) -> tuple[dict[str, torch.Tensor], torch.Tensor]:
    _dict: dict[str, torch.Tensor] = {}
    if (init or (not seq_only or isinstance(module, nn.Sequential))) \
            and depth != 0:
        for name, child in module.named_children():
            full_name = prefix + ('.' if prefix else '') + \
                name  # prefix=full_name
            if layer_input == 'record' or \
                    layer_input.startswith(f'{full_name}.'):
                sub_dict, x = _get_all_layer(child, x, layer_input, depth - 1,
                                             full_name, use_filter, non_leaf,
                                             seq_only, verbose)
                _dict.update(sub_dict)
                layer_input = 'record'
            elif layer_input == full_name:
                layer_input = 'record'
    else:
        x = module(x)
    if prefix and (not use_filter or filter_layer(module)) \
            and (non_leaf or depth == 0 or
                 not isinstance(module, nn.Sequential)):
        _dict[prefix] = x.clone()
        if verbose:
            shape_str = str(list(x.shape))
            module_str = ''
            match verbose:
                case 1:
                    module_str = module.__class__.__name__
                case 2:
                    module_str = type(module)
                case 3:
                    module_str = str(module).split('\n')[0].removesuffix('(')
                case _:
                    module_str = str(module)
            print(f'{ansi["blue_light"]}{prefix:<50s}{ansi["reset"]}'
                  f'{ansi["yellow"]}{shape_str:<20}{ansi["reset"]}'
                  f'{module_str}')
    return _dict, x


def get_layer(module: nn.Module, x: torch.Tensor, layer_output: str = 'output',
              layer_input: str = 'input',
              layer_name_list: list[str] = None,
              seq_only: bool = True) -> torch.Tensor:
    r"""Get one certain intermediate layer output
    of :attr:`_input` from any intermediate layer
    in a :any:`torch.nn.Module`.

    Args:
        module (torch.nn.Module): the module to process.
        x (torch.Tensor): The batched input tensor
            from :attr:`layer_input`.
        layer_output (str): The intermediate output layer name.
            Defaults to ``'classifier'``.
        layer_input (str): The intermediate layer name that outputs :attr:`x`.
            Defaults to ``'input'``.
        seq_only (bool): Whether to only traverse children
            of :any:`torch.nn.Sequential`.
            If ``False``, will traverse children of all :any:`torch.nn.Module`.
            Defaults to ``True``.

    Returns:
        torch.Tensor: The output of layer :attr:`layer_output`.

    :Example:
        >>> import torch
        >>> import torchvision
        >>> from trojanzoo.utils.model import get_all_layer, get_layer
        >>>
        >>> model = torchvision.models.densenet121()
        >>> x = torch.randn(5, 3, 224, 224)
        >>> y = get_all_layer(model.features, x, verbose=True)
        layer name                                        output shape        module information
        conv0                                             [5, 64, 112, 112]   Conv2d
        pool0                                             [5, 64, 56, 56]     MaxPool2d
        denseblock1                                       [5, 256, 56, 56]    _DenseBlock
        transition1.conv                                  [5, 128, 56, 56]    Conv2d
        transition1.pool                                  [5, 128, 28, 28]    AvgPool2d
        denseblock2                                       [5, 512, 28, 28]    _DenseBlock
        transition2.conv                                  [5, 256, 28, 28]    Conv2d
        transition2.pool                                  [5, 256, 14, 14]    AvgPool2d
        denseblock3                                       [5, 1024, 14, 14]   _DenseBlock
        transition3.conv                                  [5, 512, 14, 14]    Conv2d
        transition3.pool                                  [5, 512, 7, 7]      AvgPool2d
        denseblock4                                       [5, 1024, 7, 7]     _DenseBlock
        >>> x = torch.randn(6, 256, 56, 56)
        >>> get_layer(model.features, x, layer_input='denseblock1', layer_output='transition3.conv').shape
        torch.Size([6, 512, 14, 14])
    """  # noqa: E501
    if layer_input == 'input' and layer_output == 'output':
        return module(x)
    if layer_name_list is None:
        layer_name_list = get_layer_name(module, use_filter=False, non_leaf=True)
        layer_name_list.insert(0, 'input')
        layer_name_list.append('output')
    if layer_input not in layer_name_list \
        or layer_output not in layer_name_list \
            or layer_name_list.index(
                layer_input) > layer_name_list.index(layer_output):
        print('Model Layer Name List: \n', layer_name_list)
        print('Input  layer: ', layer_input)
        print('Output layer: ', layer_output)
        raise ValueError('Layer name not correct')
    if layer_input == 'input':
        layer_input = 'record'
    return _get_layer(module, x, layer_output, layer_input,
                      seq_only=seq_only, init=True)


def _get_layer(module: nn.Module, x: torch.Tensor,
               layer_output: str = 'output', layer_input: str = 'record',
               prefix: str = '', seq_only: bool = True,
               init: bool = False) -> torch.Tensor:
    if init or (not seq_only or isinstance(module, nn.Sequential)):
        for name, child in module.named_children():
            full_name = prefix + ('.' if prefix else '') + \
                name  # prefix=full_name
            if layer_input == 'record' or \
                    layer_input.startswith(f'{full_name}.'):
                x = _get_layer(child, x, layer_output, layer_input,
                               full_name, seq_only)
                layer_input = 'record'
            elif layer_input == full_name:
                layer_input = 'record'
            if layer_output == full_name or layer_output.startswith(full_name + '.'):
                return x
    else:
        x = module(x)
    return x


def filter_layer(module: nn.Module,
                 filter_tuple: tuple[nn.Module] = filter_tuple
                 ) -> bool:
    return not isinstance(module, filter_tuple)


def summary(module: nn.Module, depth: int = 0, verbose: bool = True,
            indent: int = 0, tree_length: int = None, indent_atom: int = 12
            ) -> None:
    r"""
    | Prints a string summary of the module.
    | This method is similar to `tensorflow.keras.Model.summary() <https://www.tensorflow.org/api_docs/python/tf/keras/Model#summary>`_.

    Args:
        module (torch.nn.Module): The module to process.
        depth (int): The traverse depth. Defaults to ``0``.
        verbose (bool): Whether to output auxiliary information.
            Defaults to ``True``.
        indent (int): The space indent for the entire string.
            Defaults to ``0``.
        tree_length (int): The tree length.
            If ``None``, use ``indent_atom * (depth + 1)``.
            Defaults to ``None``.
        indent_atom (int): The indent incremental for each sub-structure.
            Defaults to ``12``.

    :Example:
        >>> import torchvision
        >>> from trojanzoo.utils.model import summary
        >>>
        >>> model=torchvision.models.resnet18()
        >>> summary(model)
        >>> summary(model, depth=1)
        conv1                   Conv2d(3, 64, kernel_size=(7, 7), stride=(2, 2), padding=(3, 3), bias=False)
        bn1                     BatchNorm2d(64, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)
        relu                    ReLU(inplace=True)
        maxpool                 MaxPool2d(kernel_size=3, stride=2, padding=1, dilation=1, ceil_mode=False)
        layer1                  Sequential
        layer2                  Sequential
        layer3                  Sequential
        layer4                  Sequential
        avgpool                 AdaptiveAvgPool2d(output_size=(1, 1))
        fc                      Linear(in_features=512, out_features=1000, bias=True)
        >>> summary(model, depth=1, verbose=False)
        conv1
        bn1
        relu
        maxpool
        layer1
        layer2
        layer3
        layer4
        avgpool
        fc
        >>> summary(model, depth=2)
        conv1                               Conv2d(3, 64, kernel_size=(7, 7), stride=(2, 2), padding=(3, 3), bias=False)
        bn1                                 BatchNorm2d(64, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True)
        relu                                ReLU(inplace=True)
        maxpool                             MaxPool2d(kernel_size=3, stride=2, padding=1, dilation=1, ceil_mode=False)
        layer1                              Sequential
                    0                       BasicBlock
                    1                       BasicBlock
        layer2                              Sequential
                    0                       BasicBlock
                    1                       BasicBlock
        layer3                              Sequential
                    0                       BasicBlock
                    1                       BasicBlock
        layer4                              Sequential
                    0                       BasicBlock
                    1                       BasicBlock
        avgpool                             AdaptiveAvgPool2d(output_size=(1, 1))
        fc                                  Linear(in_features=512, out_features=1000, bias=True)

    Note:
        You could use :func:`get_all_layer` with ``verbose=True`` to see the output tensor shape for each layer.
    """  # noqa: E501
    tree_length = tree_length or indent_atom * (depth + 1)
    if depth > 0:
        for name, child in module.named_children():
            _str = f'{ansi["blue_light"]}{name}{ansi["reset"]}'
            if verbose:
                _str = _str.ljust(tree_length - indent +
                                  len(ansi['blue_light']) + len(ansi['reset']))
                _str += str(child).split('\n')[0].removesuffix('(')
            prints(_str, indent=indent)
            summary(child, depth=depth - 1, verbose=verbose,
                    indent=indent + indent_atom,
                    tree_length=tree_length,
                    indent_atom=indent_atom)


def activate_params(module: nn.Module, params: Iterator[nn.Parameter] = []):
    r"""Set ``requires_grad=True`` for selected :attr:`params` of :attr:`module`.
    All other params are frozen.

    Args:
        module (torch.nn.Module): The module to process.
        params (~collections.abc.Iterator[torch.nn.parameter.Parameter]):
            The parameters to ``requires_grad``.
                Defaults to ``[]``.
    """
    module.requires_grad_(False)
    for param in params:
        param.requires_grad_()


@torch.no_grad()
def accuracy(_output: torch.Tensor, _label: torch.Tensor, num_classes: int,
             topk: Iterable[int] = (1, 5)) -> list[float]:
    r"""Computes the accuracy over the k top predictions
    for the specified values of k.

    Args:
        _output (torch.Tensor): The batched logit tensor with shape ``(N, C)``.
        _label (torch.Tensor): The batched label tensor with shape ``(N)``.
        num_classes (int): Number of classes.
        topk (~collections.abc.Iterable[int]): Which top-k accuracies to show.
            Defaults to ``(1, 5)``.

    Returns:
        list[float]: Top-k accuracies.
    """
    maxk = min(max(topk), num_classes)
    batch_size = _label.size(0)
    _, pred = _output.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(_label[None])
    res: list[float] = []
    for k in topk:
        if k > num_classes:
            res.append(100.0)
        else:
            correct_k = float(correct[:k].sum(dtype=torch.float32))
            res.append(correct_k * (100.0 / batch_size))
    return res


@torch.no_grad()
def generate_target(module: nn.Module, _input: torch.Tensor,
                    idx: int = 1, same: bool = False
                    ) -> torch.Tensor:
    r"""Generate target labels of a batched input based on
        the classification confidence ranking index.

    Args:
        module (torch.nn.Module): The module to process.
        _input (torch.Tensor): The input tensor.
        idx (int): The classification confidence
            rank of target class.
            Defaults to ``1``.
        same (bool): Generate the same label
            for all samples using mod.
            Defaults to ``False``.

    Returns:
        torch.Tensor:
            The generated target label with shape ``(N)``.
    """
    _output: torch.Tensor = module(_input)
    target = _output.argsort(dim=-1, descending=True)[:, idx]
    if same:
        target = repeat_to_batch(target.mode(dim=0)[0], len(_input))
    return target


class ExponentialMovingAverage(torch.optim.swa_utils.AveragedModel):
    r"""Maintains moving averages of model parameters using an exponential decay.
    ``ema_avg = decay * avg_model_param + (1 - decay) * model_param``

    See Also:
        https://github.com/pytorch/vision/blob/main/references/classification/utils.py

        `torch.optim.swa_utils.AveragedModel <https://pytorch.org/docs/stable/optim.html#stochastic-weight-averaging>`_
        is used to compute the EMA.
    """  # noqa: E501

    def __init__(self, model: nn.Module, decay: float):
        def ema_avg(avg_model_param: torch.Tensor, model_param: torch.Tensor,
                    num_averaged: torch.Tensor) -> torch.Tensor:
            return decay * avg_model_param + (1 - decay) * model_param

        super().__init__(model, device=env['device'], avg_fn=ema_avg)
        self.n_averaged: torch.Tensor
        self.module: nn.Module
        self.avg_fn: Callable[[torch.Tensor, torch.Tensor, torch.Tensor],
                              torch.Tensor]

    def update_parameters(self, model: nn.Module):
        for p_swa, p_model in zip(self.module.state_dict().values(),
                                  model.state_dict().values()):
            device = p_swa.device
            p_model_ = p_model.detach().to(device)
            if self.n_averaged.eq(0):
                p_swa.detach().copy_(p_model_)
            else:
                p_swa.detach().copy_(self.avg_fn(p_swa.detach(), p_model_,
                                                 self.n_averaged.to(device)))
        self.n_averaged += 1



#!/usr/bin/env python3

from ..abstract import BackdoorDefense

from .attack_agnostic import *
from .input_filtering import *
from .model_inspection import *
from .training_filtering import *

from . import attack_agnostic, input_filtering, model_inspection, training_filtering

module_list = [attack_agnostic, input_filtering, model_inspection, training_filtering]
__all__ = []
class_dict: dict[str, type[BackdoorDefense]] = {}
for module in module_list:
    __all__.extend(module.__all__)
    class_dict.update(module.class_dict)
